import unittest
import sys
from abc import ABC, abstractmethod
from jsbgym.rewards import (
    Reward,
    AsymptoticErrorComponent,
    ErrorComponent,
    LinearErrorComponent,
    AngularAsymptoticErrorComponent,
)
import jsbgym.tests.stubs as stubs
from typing import Type


class TestReward(unittest.TestCase):
    def test_init_error_on_empty_base_reward(self):
        empty_base_reward = ()
        shaping_reward = (1, 2)

        with self.assertRaises(ValueError):
            _ = Reward(empty_base_reward, shaping_reward)

    def test_base_reward_reward(self):
        base_reward = (1, 2)
        shaping_reward = ()

        reward = Reward(base_reward, shaping_reward)

        expected_r = sum(base_reward) / len(base_reward)
        self.assertAlmostEqual(expected_r, reward.agent_reward())

    def test_base_reward_non_shaping_reward(self):
        base_reward = (1, 2)
        shaping_reward = ()

        reward = Reward(base_reward, shaping_reward)

        expected_non_shaping_r = sum(base_reward) / len(base_reward)
        self.assertAlmostEqual(expected_non_shaping_r, reward.assessment_reward())

    def test_shaping_reward_reward(self):
        base_reward = (1, 2)
        shaping_reward = (3, 4)

        reward = Reward(base_reward, shaping_reward)

        expected_r = sum(base_reward + shaping_reward) / len(
            base_reward + shaping_reward
        )
        self.assertAlmostEqual(expected_r, reward.agent_reward())

    def test_shaping_reward_non_shaping_reward(self):
        base_reward = (1, 2)
        shaping_reward = (3, 4)

        reward = Reward(base_reward, shaping_reward)

        expected_non_shaping_r = sum(base_reward) / len(base_reward)
        self.assertAlmostEqual(expected_non_shaping_r, reward.assessment_reward())

    def test_base_reward_is_not_shaping(self):
        base_reward = (1, 2)
        shaping_reward = ()

        reward = Reward(base_reward, shaping_reward)

        self.assertFalse(reward.is_shaping())

    def test_shaping_reward_is_shaping(self):
        base_reward = (1, 2)
        shaping_reward = (3, 4)

        reward = Reward(base_reward, shaping_reward)

        self.assertTrue(reward.is_shaping())


class AbstractTestErrorComponent(unittest.TestCase, ABC):
    dummy_task = stubs.FlightTaskStub()
    default_state_vars = dummy_task.state_variables
    default_property_index = 0
    default_target_index = 1
    default_target_value = 5.0
    default_State = dummy_task.State
    default_scaling_factor = 1
    default_shaping = True
    default_name = "test_component"
    extra_kwargs = dict()

    def setUp(self):
        self.COT = self.get_class_under_test()

    @abstractmethod
    def get_class_under_test(self) -> Type[ErrorComponent]:
        ...

    def get_default_perfect_state(self):
        """Returns a state where the controlled property matches the target exactly"""
        return self.dummy_task.get_state(
            self.default_target_value, self.default_target_value
        )

    def get_default_middling_state(self):
        """Returns a state where the error (value - target) is equal to the scaling factor"""
        value = self.default_target_value + self.default_scaling_factor
        return self.dummy_task.get_state(value, self.default_target_value)

    def get_default_rubbish_state(self):
        """Returns a state where the error between value and target is very very large"""
        value = sys.float_info.max
        return self.dummy_task.get_state(value, self.default_target_value)

    def get_component_target_property(
        self,
        name=default_name,
        property_index=default_property_index,
        state_vars=default_state_vars,
        target_index=default_target_index,
        scaling_factor=default_scaling_factor,
    ):
        return self.COT(
            name,
            state_vars[property_index],
            state_vars,
            state_vars[target_index],
            scaling_factor,
            **self.extra_kwargs,
        )

    def get_component_target_constant(
        self,
        name=default_name,
        property_index=default_property_index,
        state_vars=default_state_vars,
        target_value=default_target_value,
        shaping=default_shaping,
        scaling_factor=default_scaling_factor,
    ):
        return self.COT(
            name,
            state_vars[property_index],
            state_vars,
            target_value,
            shaping,
            scaling_factor,
            **self.extra_kwargs,
        )


class TestAsymptoticErrorComponent(AbstractTestErrorComponent):
    PERFECT_POTENTIAL = 1.0
    SCALING_FACTOR_POTENTIAL = 0.5
    INF_ERROR_POTENTIAL = 0.0

    def get_class_under_test(self):
        return AsymptoticErrorComponent

    def test_calculate_get_potential_perfect_state_non_terminal(self):
        target_value = 0
        actual_value = target_value
        state = self.dummy_task.get_state(actual_value, target_value)
        terminal = False

        component: AsymptoticErrorComponent = self.get_component_target_constant(
            target_value=target_value
        )

        expected_potential = self.PERFECT_POTENTIAL
        self.assertAlmostEqual(
            expected_potential, component.get_potential(state, terminal)
        )

    def test_calculate_get_potential_perfect_state_terminal(self):
        target_value = 0
        actual_value = target_value
        state = self.dummy_task.get_state(actual_value, target_value)
        shaping = True
        terminal = True

        component: AsymptoticErrorComponent = self.get_component_target_constant(
            target_value=target_value, shaping=shaping
        )

        expected_potential = component.POTENTIAL_BASED_DIFFERENCE_TERMINAL_VALUE
        self.assertAlmostEqual(
            expected_potential, component.get_potential(state, terminal)
        )

    def test_calculate_get_potential_terrible_state(self):
        target_value = 0
        actual_value = sys.float_info.max
        state = self.dummy_task.get_state(actual_value, target_value)
        terminal = False

        component: AsymptoticErrorComponent = self.get_component_target_constant(
            target_value=target_value
        )

        expected_potential = self.INF_ERROR_POTENTIAL
        self.assertAlmostEqual(
            expected_potential, component.get_potential(state, terminal)
        )

    def test_calculate_get_potential_midway_value(self):
        target_value = 0
        actual_value = self.default_scaling_factor  # should produce potential of 0.5
        state = self.dummy_task.get_state(actual_value, target_value)
        terminal = False

        component: AsymptoticErrorComponent = self.get_component_target_constant(
            target_value=target_value
        )

        expected_potential = self.SCALING_FACTOR_POTENTIAL
        self.assertAlmostEqual(
            expected_potential, component.get_potential(state, terminal)
        )

    def test_calculate_same_perfect_states_shaping_non_terminal(self):
        last_state = self.get_default_perfect_state()
        state = self.get_default_perfect_state()
        terminal = False

        component = self.get_component_target_constant(shaping=True)

        potential_diff = self.PERFECT_POTENTIAL - self.PERFECT_POTENTIAL
        self.assertAlmostEqual(
            potential_diff, component.calculate(state, last_state, terminal)
        )

    def test_calculate_same_perfect_states_shaping_terminal(self):
        last_state = self.get_default_perfect_state()
        state = self.get_default_perfect_state()
        terminal = True

        component = self.get_component_target_constant(shaping=True)

        potential_diff = (
            component.POTENTIAL_BASED_DIFFERENCE_TERMINAL_VALUE - self.PERFECT_POTENTIAL
        )
        self.assertAlmostEqual(
            potential_diff, component.calculate(state, last_state, terminal)
        )

    def test_calculate_improved_state_shaping_non_terminal(self):
        last_state = self.get_default_middling_state()
        state = self.get_default_perfect_state()
        terminal = False

        component = self.get_component_target_constant(shaping=True)

        potential_diff = self.PERFECT_POTENTIAL - self.SCALING_FACTOR_POTENTIAL
        self.assertAlmostEqual(
            potential_diff, component.calculate(state, last_state, terminal)
        )

    def test_calculate_worsening_state_non_terminal(self):
        last_state = self.get_default_perfect_state()
        state = self.get_default_rubbish_state()
        terminal = False

        component = self.get_component_target_constant()

        potential_diff = self.INF_ERROR_POTENTIAL - self.PERFECT_POTENTIAL
        self.assertAlmostEqual(
            potential_diff, component.calculate(state, last_state, terminal)
        )


class TestLinearShapingComponent(AbstractTestErrorComponent):
    PERFECT_POTENTIAL = 1.0
    MIDDLE_POTENTIAL = 0.5
    SCALING_FACTOR_POTENTIAL = 0.0

    def get_class_under_test(self):
        return LinearErrorComponent

    def get_default_middling_state(self):
        """
        Returns a state where the error (value - target) is equal to
        half of the scaling factor, corresponding to normalised error of 0.5
        """
        value = self.default_target_value - 0.5 * self.default_scaling_factor
        return self.dummy_task.get_state(value, self.default_target_value)

    def get_default_rubbish_state(self):
        """
        Returns a state where the error between value and target is at maximum,
        i.e. equal to the scaling factor.
        """
        value = self.default_target_value - self.default_scaling_factor
        return self.dummy_task.get_state(value, self.default_target_value)

    def test_calculate_get_potential_perfect_state_non_terminal(self):
        state = self.get_default_perfect_state()
        terminal = False

        component: LinearErrorComponent = self.get_component_target_constant()
        self.assertIsInstance(component, LinearErrorComponent)

        self.assertAlmostEqual(
            self.PERFECT_POTENTIAL, component.get_potential(state, terminal)
        )

    def test_calculate_get_potential_perfect_state_terminal(self):
        state = self.get_default_perfect_state()
        terminal = True

        component: LinearErrorComponent = self.get_component_target_constant()

        self.assertAlmostEqual(
            component.POTENTIAL_BASED_DIFFERENCE_TERMINAL_VALUE,
            component.get_potential(state, terminal),
        )

    def test_calculate_get_potential_rubbish_state(self):
        state = self.get_default_rubbish_state()
        terminal = False

        component: LinearErrorComponent = self.get_component_target_constant()

        self.assertAlmostEqual(
            self.SCALING_FACTOR_POTENTIAL, component.get_potential(state, terminal)
        )

    def test_calculate_get_potential_midway_value(self):
        state = self.get_default_middling_state()
        terminal = False

        component: LinearErrorComponent = self.get_component_target_constant()

        self.assertAlmostEqual(
            self.MIDDLE_POTENTIAL, component.get_potential(state, terminal)
        )

    def test_calculate_same_states_non_terminal(self):
        last_state = self.get_default_perfect_state()
        state = self.get_default_perfect_state()
        terminal = False

        component = self.get_component_target_constant()

        self.assertAlmostEqual(0.0, component.calculate(state, last_state, terminal))

    def test_calculate_same_states_terminal(self):
        last_state = self.get_default_perfect_state()
        state = self.get_default_perfect_state()
        terminal = True

        component = self.get_component_target_constant(shaping=True)

        pot_difference = (
            component.POTENTIAL_BASED_DIFFERENCE_TERMINAL_VALUE - self.PERFECT_POTENTIAL
        )
        self.assertAlmostEqual(
            pot_difference, component.calculate(state, last_state, terminal)
        )

    def test_calculate_improved_state_shaping_non_terminal(self):
        last_state = self.get_default_middling_state()
        state = self.get_default_perfect_state()
        terminal = False

        component = self.get_component_target_constant(shaping=True)

        self.assertAlmostEqual(0.5, component.calculate(state, last_state, terminal))

    def test_calculate_worsening_state_non_terminal(self):
        last_state = self.get_default_perfect_state()
        state = self.get_default_rubbish_state()
        terminal = False

        component = self.get_component_target_constant()

        self.assertAlmostEqual(-1.0, component.calculate(state, last_state, terminal))


class TestAngularAsmyptoticErrorComponent(unittest.TestCase):
    dummy_task = stubs.FlightTaskStub()
    default_state_vars = dummy_task.state_variables
    default_property_index = 0
    default_target_index = 1
    default_target_value = 90.0
    default_State = dummy_task.State
    default_scaling_factor = 10
    default_name = "test_component"

    PERFECT_POTENTIAL = 1.0
    SCALING_FACTOR_POTENTIAL = 0.5
    INF_ERROR_POTENTIAL = 0.0

    @staticmethod
    def get_component_constant(
        name=default_name,
        property_index=default_property_index,
        state_vars=default_state_vars,
        target_value=default_target_value,
        scaling_factor=default_scaling_factor,
    ) -> AngularAsymptoticErrorComponent:
        return AngularAsymptoticErrorComponent(
            name=name,
            prop=state_vars[property_index],
            state_variables=state_vars,
            target=target_value,
            is_potential_based=True,
            scaling_factor=scaling_factor,
        )

    def test_get_potential_value_equals_target_terminal(self):
        value = self.default_target_value
        target = self.default_target_value
        state = self.dummy_task.get_state(value, target)
        terminal = True

        component = self.get_component_constant()

        self.assertAlmostEqual(0.0, component.get_potential(state, terminal))

    def test_get_potential_value_equals_target_non_terminal(self):
        value = self.default_target_value
        target = self.default_target_value
        state = self.dummy_task.get_state(value, target)
        terminal = False

        component = self.get_component_constant()

        self.assertAlmostEqual(
            self.PERFECT_POTENTIAL, component.get_potential(state, terminal)
        )

    def test_get_potential_value_scaling_factor_off_target(self):
        target = self.default_target_value
        value_left = target - self.default_scaling_factor
        value_right = target + self.default_scaling_factor
        state_left_of_target = self.dummy_task.get_state(value_left, target)
        state_right_of_target = self.dummy_task.get_state(value_right, target)
        terminal = False

        component = self.get_component_constant()

        self.assertAlmostEqual(
            self.SCALING_FACTOR_POTENTIAL,
            component.get_potential(state_left_of_target, terminal),
        )
        self.assertAlmostEqual(
            self.SCALING_FACTOR_POTENTIAL,
            component.get_potential(state_right_of_target, terminal),
        )

    def test_get_potential_equal_across_360_degrees(self):
        target = 0
        value_left_negative = target - 1.5
        value_left_positive = 360 - target - 1.5  # equivalent
        value_right = target + 1.5
        state_left_of_target_negative = self.dummy_task.get_state(
            value_left_negative, target
        )
        state_left_of_target_positive = self.dummy_task.get_state(
            value_left_positive, target
        )
        state_right_of_target = self.dummy_task.get_state(value_right, target)
        terminal = False

        component = self.get_component_constant(target_value=target)
        potential_left_negative = component.get_potential(
            state_left_of_target_negative, terminal
        )
        potential_left_positive = component.get_potential(
            state_left_of_target_positive, terminal
        )
        potential_right = component.get_potential(state_right_of_target, terminal)

        self.assertLess(potential_left_negative, self.PERFECT_POTENTIAL)
        self.assertLess(potential_left_positive, self.PERFECT_POTENTIAL)
        self.assertLess(potential_right, self.PERFECT_POTENTIAL)
        self.assertAlmostEqual(potential_left_negative, potential_left_positive)
        self.assertAlmostEqual(potential_left_negative, potential_right)



import math
import collections
from jsbgym import utils


class BoundedProperty(
    collections.namedtuple("BoundedProperty", ["name", "description", "min", "max"])
):
    def get_legal_name(self):
        return utils.AttributeFormatter.translate(self.name)


class Property(collections.namedtuple("Property", ["name", "description"])):
    def get_legal_name(self):
        return utils.AttributeFormatter.translate(self.name)


# position and attitude
altitude_sl_ft = BoundedProperty(
    "position/h-sl-ft", "altitude above mean sea level [ft]", -1400, 85000
)
pitch_rad = BoundedProperty(
    "attitude/pitch-rad", "pitch [rad]", -0.5 * math.pi, 0.5 * math.pi
)
roll_rad = BoundedProperty("attitude/roll-rad", "roll [rad]", -math.pi, math.pi)
heading_deg = BoundedProperty("attitude/psi-deg", "heading [deg]", 0, 360)
sideslip_deg = BoundedProperty("aero/beta-deg", "sideslip [deg]", -180, +180)
lat_geod_deg = BoundedProperty(
    "position/lat-geod-deg", "geocentric latitude [deg]", -90, 90
)
lng_geoc_deg = BoundedProperty(
    "position/long-gc-deg", "geodesic longitude [deg]", -180, 180
)
dist_travel_m = Property(
    "position/distance-from-start-mag-mt",
    "distance travelled from starting position [m]",
)

# velocities
u_fps = BoundedProperty(
    "velocities/u-fps", "body frame x-axis velocity [ft/s]", -2200, 2200
)
v_fps = BoundedProperty(
    "velocities/v-fps", "body frame y-axis velocity [ft/s]", -2200, 2200
)
w_fps = BoundedProperty(
    "velocities/w-fps", "body frame z-axis velocity [ft/s]", -2200, 2200
)
v_north_fps = BoundedProperty(
    "velocities/v-north-fps", "velocity true north [ft/s]", float("-inf"), float("+inf")
)
v_east_fps = BoundedProperty(
    "velocities/v-east-fps", "velocity east [ft/s]", float("-inf"), float("+inf")
)
v_down_fps = BoundedProperty(
    "velocities/v-down-fps", "velocity downwards [ft/s]", float("-inf"), float("+inf")
)
p_radps = BoundedProperty(
    "velocities/p-rad_sec", "roll rate [rad/s]", -2 * math.pi, 2 * math.pi
)
q_radps = BoundedProperty(
    "velocities/q-rad_sec", "pitch rate [rad/s]", -2 * math.pi, 2 * math.pi
)
r_radps = BoundedProperty(
    "velocities/r-rad_sec", "yaw rate [rad/s]", -2 * math.pi, 2 * math.pi
)
altitude_rate_fps = Property("velocities/h-dot-fps", "Rate of altitude change [ft/s]")

# controls state
aileron_left = BoundedProperty(
    "fcs/left-aileron-pos-norm", "left aileron position, normalised", -1, 1
)
aileron_right = BoundedProperty(
    "fcs/right-aileron-pos-norm", "right aileron position, normalised", -1, 1
)
elevator = BoundedProperty(
    "fcs/elevator-pos-norm", "elevator position, normalised", -1, 1
)
rudder = BoundedProperty("fcs/rudder-pos-norm", "rudder position, normalised", -1, 1)
throttle = BoundedProperty(
    "fcs/throttle-pos-norm", "throttle position, normalised", 0, 1
)
gear = BoundedProperty("gear/gear-pos-norm", "landing gear position, normalised", 0, 1)

# engines
engine_running = Property("propulsion/engine/set-running", "engine running (0/1 bool)")
all_engine_running = Property(
    "propulsion/set-running", "set engine running (-1 for all engines)"
)
engine_thrust_lbs = Property("propulsion/engine/thrust-lbs", "engine thrust [lb]")

# controls command
aileron_cmd = BoundedProperty(
    "fcs/aileron-cmd-norm", "aileron commanded position, normalised", -1.0, 1.0
)
elevator_cmd = BoundedProperty(
    "fcs/elevator-cmd-norm", "elevator commanded position, normalised", -1.0, 1.0
)
rudder_cmd = BoundedProperty(
    "fcs/rudder-cmd-norm", "rudder commanded position, normalised", -1.0, 1.0
)
throttle_cmd = BoundedProperty(
    "fcs/throttle-cmd-norm", "throttle commanded position, normalised", 0.0, 1.0
)
mixture_cmd = BoundedProperty(
    "fcs/mixture-cmd-norm", "engine mixture setting, normalised", 0.0, 1.0
)
throttle_1_cmd = BoundedProperty(
    "fcs/throttle-cmd-norm[1]", "throttle 1 commanded position, normalised", 0.0, 1.0
)
mixture_1_cmd = BoundedProperty(
    "fcs/mixture-cmd-norm[1]", "engine mixture 1 setting, normalised", 0.0, 1.0
)
gear_all_cmd = BoundedProperty(
    "gear/gear-cmd-norm", "all landing gear commanded position, normalised", 0, 1
)

# simulation
sim_dt = Property("simulation/dt", "JSBSim simulation timestep [s]")
sim_time_s = Property("simulation/sim-time-sec", "Simulation time [s]")

# initial conditions
initial_altitude_ft = Property("ic/h-sl-ft", "initial altitude MSL [ft]")
initial_terrain_altitude_ft = Property(
    "ic/terrain-elevation-ft", "initial terrain alt [ft]"
)
initial_longitude_geoc_deg = Property(
    "ic/long-gc-deg", "initial geocentric longitude [deg]"
)
initial_latitude_geod_deg = Property(
    "ic/lat-geod-deg", "initial geodesic latitude [deg]"
)
initial_u_fps = Property(
    "ic/u-fps", "body frame x-axis velocity; positive forward [ft/s]"
)
initial_v_fps = Property(
    "ic/v-fps", "body frame y-axis velocity; positive right [ft/s]"
)
initial_w_fps = Property("ic/w-fps", "body frame z-axis velocity; positive down [ft/s]")
initial_p_radps = Property("ic/p-rad_sec", "roll rate [rad/s]")
initial_q_radps = Property("ic/q-rad_sec", "pitch rate [rad/s]")
initial_r_radps = Property("ic/r-rad_sec", "yaw rate [rad/s]")
initial_roc_fpm = Property("ic/roc-fpm", "initial rate of climb [ft/min]")
initial_heading_deg = Property("ic/psi-true-deg", "initial (true) heading [deg]")


class Vector2(object):
    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def heading_deg(self):
        """Calculate heading in degrees of vector from origin"""
        heading_rad = math.atan2(self.x, self.y)
        heading_deg_normalised = (math.degrees(heading_rad) + 360) % 360
        return heading_deg_normalised

    @staticmethod
    def from_sim(sim: "simulation.Simulation") -> "Vector2":
        return Vector2(sim[v_east_fps], sim[v_north_fps])


class GeodeticPosition(object):
    def __init__(self, latitude_deg: float, longitude_deg: float):
        self.lat = latitude_deg
        self.lon = longitude_deg

    def heading_deg_to(self, destination: "GeodeticPosition") -> float:
        """Determines heading in degrees of course between self and destination"""
        difference_vector = destination - self
        return difference_vector.heading_deg()

    @staticmethod
    def from_sim(sim: "simulation.Simulation") -> "GeodeticPosition":
        """Return a GeodeticPosition object with lat and lon from simulation"""
        lat_deg = sim[lat_geod_deg]
        lon_deg = sim[lng_geoc_deg]
        return GeodeticPosition(lat_deg, lon_deg)

    def __sub__(self, other) -> Vector2:
        """Returns difference between two coords as (delta_lat, delta_long)"""
        return Vector2(self.lon - other.lon, self.lat - other.lat)



import gymnasium as gym
import subprocess
import time
import os
import matplotlib as mpt
import matplotlib.pyplot as plt
import jsbgym.properties as prp
from jsbgym.aircraft import Aircraft
from jsbgym.simulation import Simulation
from typing import NamedTuple, Tuple


class AxesTuple(NamedTuple):
    """Holds references to figure subplots (axes)"""

    axes_state: plt.Axes
    axes_stick: plt.Axes
    axes_throttle: plt.Axes
    axes_rudder: plt.Axes


class FigureVisualiser(object):
    """Class for manging a matplotlib Figure displaying agent state and actions"""

    PLOT_PAUSE_SECONDS = 0.0001
    LABEL_TEXT_KWARGS = dict(
        fontsize=18, horizontalalignment="right", verticalalignment="baseline"
    )
    VALUE_TEXT_KWARGS = dict(
        fontsize=18, horizontalalignment="left", verticalalignment="baseline"
    )
    TEXT_X_POSN_LABEL = 0.8
    TEXT_X_POSN_VALUE = 0.9
    TEXT_Y_POSN_INITIAL = 1.0
    TEXT_Y_INCREMENT = -0.1

    def __init__(self, _: Simulation, print_props: Tuple[prp.Property]):
        """
        Constructor.
        Sets here is ft_per_deg_lon, which depends dynamically on aircraft's
        longitude (because of the conversion between geographic and Euclidean
        coordinate systems). We retrieve longitude from the simulation and
        assume it is constant thereafter.
        :param _: (unused) Simulation that will be plotted
        :param print_props: Propertys which will have their values printed to Figure.
            Must be retrievable from the plotted Simulation.
        """
        self.print_props = print_props
        self.figure: plt.Figure = None
        self.axes: AxesTuple = None
        self.value_texts: Tuple[plt.Text] = None

    def plot(self, sim: Simulation) -> None:
        """
        Creates or updates a 3D plot of the episode.
        :param sim: Simulation that will be plotted
        """
        mpt.use("TkAgg")
        if not self.figure:
            self.figure, self.axes = self._plot_configure()

        # delete old control surface data points
        for subplot in self.axes[1:]:
            # pop and translate all data points
            while subplot.lines:
                data = subplot.lines[0].remove()
                del data

        self._print_state(sim)
        self._plot_control_states(sim, self.axes)
        self._plot_control_commands(sim, self.axes)
        plt.pause(self.PLOT_PAUSE_SECONDS)  # voodoo pause needed for figure to update

    def close(self):
        if self.figure:
            plt.close(self.figure)
            self.figure = None
            self.axes = None

    def _plot_configure(self):
        """
        Creates a figure with subplots for states and actions.
        :return: (figure, axes) where:
            figure: a matplotlib Figure with subplots for state and controls
            axes: an AxesTuple object with references to all figure subplot axes
        """
        plt.ion()  # interactive mode allows dynamic updating of plot
        figure = plt.figure(figsize=(6, 11))

        spec = plt.GridSpec(
            nrows=3,
            ncols=2,
            width_ratios=[5, 1],  # second column very thin
            height_ratios=[6, 5, 1],  # bottom row very short
            wspace=0.3,
        )

        # create subplots
        axes_state = figure.add_subplot(spec[0, 0:])
        axes_stick = figure.add_subplot(spec[1, 0])
        axes_throttle = figure.add_subplot(spec[1, 1])
        axes_rudder = figure.add_subplot(spec[2, 0])

        # hide state subplot axes - text will be printed to it
        axes_state.axis("off")
        self._prepare_state_printing(axes_state)

        # config subplot for stick (aileron and elevator control in x/y axes)
        axes_stick.set_xlabel(
            "ailerons [-]",
        )
        axes_stick.set_ylabel("elevator [-]")
        axes_stick.set_xlim(left=-1, right=1)
        axes_stick.set_ylim(bottom=-1, top=1)
        axes_stick.xaxis.set_label_coords(0.5, 1.08)
        axes_stick.yaxis.set_label_coords(-0.05, 0.5)
        # make axes cross at origin
        axes_stick.spines["left"].set_position("zero")
        axes_stick.spines["bottom"].set_position("zero")
        # only show ticks at extremes of range
        axes_stick.set_xticks([-1, 1])
        axes_stick.xaxis.set_ticks_position("bottom")
        axes_stick.set_yticks([-1, 1])
        axes_stick.yaxis.set_ticks_position("left")
        axes_stick.tick_params(which="both", direction="inout")
        # show minor ticks throughout
        minor_locator = plt.MultipleLocator(0.2)
        axes_stick.xaxis.set_minor_locator(minor_locator)
        axes_stick.yaxis.set_minor_locator(minor_locator)
        # hide unneeded spines
        axes_stick.spines["right"].set_visible(False)
        axes_stick.spines["top"].set_visible(False)

        # config subplot for throttle: a 1D vertical plot
        axes_throttle.set_ylabel("throttle [-]")
        axes_throttle.set_ylim(bottom=0, top=1)
        axes_throttle.set_xlim(left=0, right=1)
        axes_throttle.spines["left"].set_position("zero")
        axes_throttle.yaxis.set_label_coords(0.5, 0.5)
        axes_throttle.set_yticks([0, 0.5, 1])
        axes_throttle.yaxis.set_minor_locator(minor_locator)
        axes_throttle.tick_params(axis="y", which="both", direction="inout")
        # hide horizontal x-axis and related spines
        axes_throttle.xaxis.set_visible(False)
        for spine in ["right", "bottom", "top"]:
            axes_throttle.spines[spine].set_visible(False)

        # config rudder subplot: 1D horizontal plot
        axes_rudder.set_xlabel("rudder [-]")
        axes_rudder.set_xlim(left=-1, right=1)
        axes_rudder.set_ylim(bottom=0, top=1)
        axes_rudder.xaxis.set_label_coords(0.5, -0.5)
        axes_stick.spines["bottom"].set_position("zero")
        axes_rudder.set_xticks([-1, 0, 1])
        axes_rudder.xaxis.set_minor_locator(minor_locator)
        axes_rudder.tick_params(axis="x", which="both", direction="inout")
        axes_rudder.get_yaxis().set_visible(False)  # only want a 1D subplot
        for spine in ["left", "right", "top"]:
            axes_rudder.spines[spine].set_visible(False)

        all_axes = AxesTuple(
            axes_state=axes_state,
            axes_stick=axes_stick,
            axes_throttle=axes_throttle,
            axes_rudder=axes_rudder,
        )

        # create figure-wide legend
        cmd_entry = (
            plt.Line2D(
                [], [], color="b", marker="o", ms=10, linestyle="", fillstyle="none"
            ),
            "Commanded Position, normalised",
        )
        pos_entry = (
            plt.Line2D([], [], color="r", marker="+", ms=10, linestyle=""),
            "Current Position, normalised",
        )
        figure.legend(
            (cmd_entry[0], pos_entry[0]),
            (cmd_entry[1], pos_entry[1]),
            loc="lower center",
        )

        plt.show()
        plt.pause(self.PLOT_PAUSE_SECONDS)  # voodoo pause needed for figure to appear

        return figure, all_axes

    def _prepare_state_printing(self, ax: plt.Axes):
        ys = [
            self.TEXT_Y_POSN_INITIAL + i * self.TEXT_Y_INCREMENT
            for i in range(len(self.print_props))
        ]

        for prop, y in zip(self.print_props, ys):
            label = str(prop.name)
            ax.text(
                self.TEXT_X_POSN_LABEL,
                y,
                label,
                transform=ax.transAxes,
                **(self.LABEL_TEXT_KWARGS),
            )

        # print and store empty Text objects which we will rewrite each plot call
        value_texts = []
        dummy_msg = ""
        for y in ys:
            text = ax.text(
                self.TEXT_X_POSN_VALUE,
                y,
                dummy_msg,
                transform=ax.transAxes,
                **(self.VALUE_TEXT_KWARGS),
            )
            value_texts.append(text)
        self.value_texts = tuple(value_texts)

    def _print_state(self, sim: Simulation):
        # update each Text object with latest value
        for prop, text in zip(self.print_props, self.value_texts):
            text.set_text(f"{sim[prop]:.4g}")

    def _plot_control_states(self, sim: Simulation, all_axes: AxesTuple):
        control_surfaces = [prp.aileron_left, prp.elevator, prp.throttle, prp.rudder]
        ail, ele, thr, rud = [sim[control] for control in control_surfaces]
        # plot aircraft control surface positions
        all_axes.axes_stick.plot(
            [ail], [ele], "r+", mfc="none", markersize=10, clip_on=False
        )
        all_axes.axes_throttle.plot(
            [0], [thr], "r+", mfc="none", markersize=10, clip_on=False
        )
        all_axes.axes_rudder.plot(
            [rud], [0], "r+", mfc="none", markersize=10, clip_on=False
        )

    def _plot_control_commands(self, sim: Simulation, all_axes: AxesTuple):
        """
        Plots agent-commanded actions on the environment figure.
        :param sim: Simulation to plot control commands from
        :param all_axes: AxesTuple, collection of axes of subplots to plot on
        """
        ail_cmd = sim[prp.aileron_cmd]
        ele_cmd = sim[prp.elevator_cmd]
        thr_cmd = sim[prp.throttle_cmd]
        rud_cmd = sim[prp.rudder_cmd]

        all_axes.axes_stick.plot(
            [ail_cmd], [ele_cmd], "bo", mfc="none", markersize=10, clip_on=False
        )
        all_axes.axes_throttle.plot(
            [0], [thr_cmd], "bo", mfc="none", markersize=10, clip_on=False
        )
        all_axes.axes_rudder.plot(
            [rud_cmd], [0], "bo", mfc="none", markersize=10, clip_on=False
        )


class FlightGearVisualiser(object):
    """
    Class for visualising aircraft using the FlightGear simulator.

    This visualiser launches FlightGear and (by default) waits for it to
    launch. A Figure is also displayed (by creating its own FigureVisualiser)
    which is used to display the agent's actions.
    """

    TYPE = "socket"
    DIRECTION = "in"
    RATE = 60
    SERVER = ""
    PORT = 5550
    PROTOCOL = "udp"
    LOADED_MESSAGE = "loading cities done"
    LOADED_MESSAGE1 = "Starting hard-coded terrain presampling"
    FLIGHTGEAR_TIME_FACTOR = 1  # sim speed relative to realtime, higher is faster
    TIME = "morning"

    def __init__(
        self, sim: Simulation, print_props: Tuple[prp.Property], block_until_loaded=True
    ):
        """
        Launches FlightGear in subprocess and starts figure for plotting actions.

        :param sim: Simulation that will be visualised
        :param aircraft: Aircraft to be loaded in FlightGear for visualisation
        :param print_props: collection of Propertys to be printed to Figure
        :param block_until_loaded: visualiser will block until it detects that
            FlightGear has loaded if True.
        """
        self.configure_simulation_output(sim)
        self.print_props = print_props
        self.flightgear_process = self._launch_flightgear(sim.get_aircraft())
        self.figure = FigureVisualiser(sim, print_props)
        if block_until_loaded:
            self._block_until_flightgear_loaded()

    def plot(self, sim: Simulation) -> None:
        """
        Updates a 3D plot of agent actions.
        """
        self.figure.plot(sim)

    @staticmethod
    def _launch_flightgear(aircraft: Aircraft):
        cmd_line_args = FlightGearVisualiser._create_cmd_line_args(
            aircraft.flightgear_id
        )
        gym.logger.info(f'Subprocess: "{cmd_line_args}"')
        flightgear_process = subprocess.Popen(
            cmd_line_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        gym.logger.info("Started FlightGear")
        return flightgear_process

    def configure_simulation_output(self, sim: Simulation):
        sim.enable_flightgear_output()
        sim.set_simulation_time_factor(self.FLIGHTGEAR_TIME_FACTOR)

    @staticmethod
    def _create_cmd_line_args(aircraft_id: str):
        # FlightGear doesn't have a 172X model, use the P instead
        if aircraft_id == "c172x":
            aircraft_id = "c172p"

        flightgear_cmd = "fgfs"
        aircraft_arg = f"--aircraft={aircraft_id}"
        flight_model_arg = (
            "--native-fdm=" + f"{FlightGearVisualiser.TYPE},"
            f"{FlightGearVisualiser.DIRECTION},"
            f"{FlightGearVisualiser.RATE},"
            f"{FlightGearVisualiser.SERVER},"
            f"{FlightGearVisualiser.PORT},"
            f"{FlightGearVisualiser.PROTOCOL}"
        )
        flight_model_type_arg = "--fdm=" + "external"
        disable_ai_arg = "--disable-ai-traffic"
        disable_live_weather_arg = "--disable-real-weather-fetch"
        time_of_day_arg = "--timeofday=" + FlightGearVisualiser.TIME
        return (
            flightgear_cmd,
            aircraft_arg,
            flight_model_arg,
            flight_model_type_arg,
            disable_ai_arg,
            disable_live_weather_arg,
            time_of_day_arg,
        )

    def _block_until_flightgear_loaded(self):
        while True:
            msg_out = self.flightgear_process.stdout.readline().decode()
            if self.LOADED_MESSAGE in msg_out or self.LOADED_MESSAGE1 in msg_out:
                time.sleep(5)
                print("FlightGear Loading Complete")
                break
            else:
                time.sleep(0.1)

    def close(self):
        if self.flightgear_process:
            self.flightgear_process.kill()



import time
import unittest
from jsbgym.environment import JsbSimEnv
from jsbgym.visualiser import FigureVisualiser, FlightGearVisualiser
from jsbgym.tests.stubs import BasicFlightTask, DefaultSimStub
import matplotlib.pyplot as plt
import jsbgym.visualiser


class TestFigureVisualiser(unittest.TestCase):
    sim = None
    visualiser = None

    def setUp(self, plot_position=True):
        self.sim = DefaultSimStub()
        task = BasicFlightTask()
        self.visualiser = FigureVisualiser(DefaultSimStub(), task.get_props_to_output())

    def tearDown(self):
        self.visualiser.close()

    def test_plot_creates_figure_and_axes(self):
        self.setUp()

        self.visualiser.plot(self.sim)

        self.assertIsInstance(self.visualiser.figure, plt.Figure)
        self.assertIsInstance(self.visualiser.axes, jsbgym.visualiser.AxesTuple)

    def test_plot_doesnt_plot_position_when_set_by_init(self):
        self.setUp(plot_position=False)

        self.visualiser.plot(self.sim)

        position_axis = self.visualiser.axes.axes_state
        is_empty_plot = position_axis is None or len(position_axis.axes.lines) == 0
        self.assertTrue(is_empty_plot)

    def test_plot_plots_control_state(self):
        self.setUp()

        self.visualiser.plot(self.sim)

    def test_close_removes_figure(self):
        self.setUp()
        self.visualiser.plot(self.sim)

        self.visualiser.close()

        self.assertIsNone(self.visualiser.figure)
        self.assertIsNone(self.visualiser.axes)


class TestFlightGearVisualiser(unittest.TestCase):
    env = None
    sim = None
    flightgear = None

    def setUp(self):
        if self.env:
            self.env.close()
        if self.sim:
            self.sim.close()
        self.task = BasicFlightTask()
        self.env = JsbSimEnv(task_type=BasicFlightTask)
        self.env.reset()
        self.sim = self.env.sim
        self.flightgear = None
        # individual test methods should init as needed:
        # self.flightgear = FlightGearVisualiser(self.sim)

    def tearDown(self):
        if self.env:
            self.env.close()
        if self.sim:
            self.sim.close()
        if self.flightgear:
            self.flightgear.close()

    def test_init_creates_figure(self):
        self.flightgear = FlightGearVisualiser(
            self.sim, self.task.get_props_to_output(), block_until_loaded=False
        )
        self.assertIsInstance(self.flightgear.figure, FigureVisualiser)

    def test_launch_flightgear(self):
        self.flightgear = FlightGearVisualiser(
            self.sim, self.task.get_props_to_output(), block_until_loaded=False
        )
        time.sleep(0.5)

        # check FlightGear has launched by looking at stdout
        self.assertIn(
            "FlightGear", self.flightgear.flightgear_process.stdout.readline().decode()
        )
        self.flightgear.close()

    def test_close_closes_flightgear(self):
        self.flightgear = FlightGearVisualiser(
            self.sim, self.task.get_props_to_output(), block_until_loaded=False
        )
        self.flightgear.close()
        timeout_seconds = 2.0
        return_code = self.flightgear.flightgear_process.wait(timeout=timeout_seconds)
        # a non-None return code indicates termination
        self.assertIsNotNone(return_code)

    def test_plot_displays_actions(self):
        self.setUp()
        self.flightgear = FlightGearVisualiser(
            self.sim, self.task.get_props_to_output(), block_until_loaded=False
        )
        self.flightgear.plot(self.sim)

        # the figure should have plotted a Lines object each axis
        for axis in ["axes_stick", "axes_rudder", "axes_throttle"]:
            axis_data_plots = getattr(self.flightgear.figure.axes, axis)
            is_empty_plot = len(axis_data_plots.axes.lines) == 0
            self.assertFalse(is_empty_plot, msg=f"no data plotted on axis {axis}")



import gymnasium as gym
import numpy as np
from typing import List
from abc import ABC, abstractmethod


class Agent(ABC):
    @abstractmethod
    def act(self, state) -> np.ndarray:
        ...

    @abstractmethod
    def observe(self, state, action, reward, terminated):
        ...


class RandomAgent(Agent):
    """An agent that selects random actions.

    The Random object making selection is gym.np_random used by the
    Space.sample() method. Its seed is set by gym.
    """

    def __init__(self, action_space: gym.Space):
        self.action_space = action_space

    def act(self, _):
        return self.action_space.sample()

    def observe(self, state, action, reward, terminated):
        # this agent type does not learn in response to observations
        pass


class ConstantAgent(Agent):
    def __init__(self, action_space: gym.spaces.Box):
        self.constant_action = (action_space.low + action_space.high) / 2

    def act(self, _):
        return self.constant_action

    def observe(self, state, action, reward, terminated):
        # this agent type does not learn in response to observations
        pass



import unittest
import jsbsim
import multiprocessing
import time
from jsbgym.simulation import Simulation
from jsbgym import aircraft
import jsbgym.properties as prp


class TestSimulation(unittest.TestCase):
    sim: Simulation = None

    def setUp(self):
        if self.sim:
            self.sim.close()
        self.sim = Simulation()

    def tearDown(self):
        self.sim = None

    def test_init_jsbsim(self):
        self.assertIsInstance(
            self.sim.jsbsim,
            jsbsim.FGFDMExec,
            msg=f"Expected Simulation.jsbsim to hold an " "instance of JSBSim.",
        )

    def test_load_model(self):
        plane = aircraft.a320
        self.sim = None
        self.sim = Simulation(aircraft=plane)
        actual_name = self.sim.get_loaded_model_name()

        self.assertEqual(
            plane.jsbsim_id,
            actual_name,
            msg=f"Unexpected aircraft model name after loading.",
        )

    def test_load_bad_aircraft_id(self):
        bad_name = "qwertyuiop"
        bad_aircraft = aircraft.Aircraft(bad_name, "", "", 100.0)

        with self.assertRaises(RuntimeError):
            self.sim = None
            self.sim = Simulation(aircraft=bad_aircraft)

    def test_get_property(self):
        self.setUp()
        # we expect certain values specified in the IC config XML file
        expected_values = {
            prp.initial_u_fps: 328.0,
            prp.initial_v_fps: 0.0,
            prp.initial_w_fps: 0.0,
            prp.u_fps: 328.0,
            prp.v_fps: 0.0,
            prp.w_fps: 0.0,
        }

        for prop, expected in expected_values.items():
            actual = self.sim[prop]
            self.assertAlmostEqual(expected, actual)

    def test_get_bad_property(self):
        self.setUp()
        bad_prop = prp.BoundedProperty("bad_prop_name", "", 0, 0)
        with self.assertRaises(KeyError):
            _ = self.sim[bad_prop]

    def test_set_property(self):
        self.setUp()
        set_values = {
            prp.altitude_sl_ft: 1000,
            prp.aileron_cmd: 0.2,
            prp.elevator_cmd: 0.2,
            prp.rudder_cmd: 0.2,
            prp.throttle_cmd: 0.2,
        }

        for prop, value in set_values.items():
            self.sim[prop] = value

        for prop, expected in set_values.items():
            actual = self.sim[prop]
            self.assertAlmostEqual(expected, actual)

    def test_initialise_conditions_basic_config(self):
        plane = aircraft.f15

        # manually reset JSBSim instance with new initial conditions
        if self.sim:
            self.sim.close()
        sim_frequency = 2
        self.sim = Simulation(
            sim_frequency_hz=sim_frequency, aircraft=plane, init_conditions=None
        )

        self.assertEqual(
            self.sim.get_loaded_model_name(),
            plane.jsbsim_id,
            msg="JSBSim did not load expected aircraft model: "
            + self.sim.get_loaded_model_name(),
        )

        # check that properties are as we expected them to be
        expected_values = {
            prp.initial_u_fps: 328.0,
            prp.initial_v_fps: 0.0,
            prp.initial_w_fps: 0.0,
            prp.u_fps: 328.0,
            prp.v_fps: 0.0,
            prp.w_fps: 0.0,
            prp.BoundedProperty("simulation/dt", "", None, None): 1 / sim_frequency,
        }

        for prop, expected in expected_values.items():
            actual = self.sim[prop]
            self.assertAlmostEqual(expected, actual)

    def test_initialise_conditions_custom_config(self):
        """Test JSBSimInstance initialisation with custom initial conditions."""

        plane = aircraft.f15
        init_conditions = {
            prp.initial_u_fps: 1000.0,
            prp.initial_v_fps: 0.0,
            prp.initial_w_fps: 1.0,
            prp.initial_altitude_ft: 5000,
            prp.initial_heading_deg: 12,
            prp.initial_r_radps: -0.1,
        }
        # map JSBSim initial condition properties to sim properties
        init_to_sim_conditions = {
            prp.initial_u_fps: prp.u_fps,
            prp.initial_v_fps: prp.v_fps,
            prp.initial_w_fps: prp.w_fps,
            prp.initial_altitude_ft: prp.altitude_sl_ft,
            prp.initial_heading_deg: prp.heading_deg,
            prp.initial_r_radps: prp.r_radps,
        }
        sim_frequency = 10

        # manually reset JSBSim instance
        if self.sim:
            self.sim.close()
        self.sim = Simulation(sim_frequency, plane, init_conditions)

        # check JSBSim initial condition and simulation properties
        for init_prop, expected in init_conditions.items():
            sim_prop = init_to_sim_conditions[init_prop]

            init_actual = self.sim[init_prop]
            sim_actual = self.sim[sim_prop]
            self.assertAlmostEqual(
                expected, init_actual, msg=f"wrong value for property {init_prop}"
            )
            self.assertAlmostEqual(
                expected, sim_actual, msg=f"wrong value for property {sim_prop}"
            )

        self.assertAlmostEqual(1.0 / sim_frequency, self.sim[prp.sim_dt])

    def test_multiprocess_simulations(self):
        """
        JSBSim segfaults when multiple instances are run on one process.

        Let's confirm that we can launch multiple processes each with 1 instance.
        """
        processes = 4
        with multiprocessing.Pool(processes) as pool:
            # N.B. basic_task is a top level function that inits JSBSim
            future_results = [pool.apply_async(basic_task) for _ in range(processes)]
            results = [f.get() for f in future_results]

        good_exit_code = 0
        expected = [good_exit_code] * processes
        self.assertListEqual(
            results, expected, msg="multiprocess execution of JSBSim failed"
        )


def basic_task():
    """A simple task involving initing a JSBSimInstance to test multiprocessing."""
    model = aircraft.cessna172P
    time.sleep(0.05)
    fdm = Simulation(aircraft=model)
    fdm.run()
    time.sleep(0.05)

    return 0


if __name__ == "__main__":
    unittest.main()



import jsbgym.properties as prp
from abc import ABC, abstractmethod
from typing import Tuple, Union
from jsbgym.utils import reduce_reflex_angle_deg

State = "tasks.FlightTask.State"  # alias for type hint


class Reward(object):
    """
    Immutable class storing an RL reward.

    We decompose rewards into tuples of component values, reflecting contributions
    from different goals. Separate tuples are maintained for the assessment (non-shaping)
    components and the shaping components. It is intended that the

    Scalar reward values are retrieved by calling .reward() or non_shaping_reward().
    The scalar value is the mean of the components.
    """

    def __init__(self, base_reward_elements: Tuple, shaping_reward_elements: Tuple):
        self.base_reward_elements = base_reward_elements
        self.shaping_reward_elements = shaping_reward_elements
        if not self.base_reward_elements:
            raise ValueError("base agent_reward cannot be empty")

    def agent_reward(self) -> float:
        """Returns scalar reward value by taking mean of all reward elements"""
        sum_reward = sum(self.base_reward_elements) + sum(self.shaping_reward_elements)
        num_reward_components = len(self.base_reward_elements) + len(
            self.shaping_reward_elements
        )
        return sum_reward / num_reward_components

    def assessment_reward(self) -> float:
        """Returns scalar non-shaping reward by taking mean of base reward elements."""
        return sum(self.base_reward_elements) / len(self.base_reward_elements)

    def is_shaping(self):
        return bool(self.shaping_reward_elements)


class RewardComponent(ABC):
    """Interface for RewardComponent, an object which calculates one component value of a Reward"""

    @abstractmethod
    def calculate(self, state: State, last_state: State, is_terminal: bool) -> float:
        ...

    @abstractmethod
    def get_name(self) -> str:
        ...

    @abstractmethod
    def get_potential(self, state: State, is_terminal) -> float:
        ...

    @abstractmethod
    def is_potential_difference_based(self) -> bool:
        ...


class NormalisedComponent(RewardComponent, ABC):
    """
    Base implementation of a RewardComponent implementing common methods.

    All potentials of subclasses should be normalised in [0.0, 1.0]
    """

    POTENTIAL_BASED_DIFFERENCE_TERMINAL_VALUE = 0.0

    def __init__(
        self,
        name: str,
        prop: prp.BoundedProperty,
        state_variables: Tuple[prp.BoundedProperty],
        target: Union[int, float, prp.Property, prp.BoundedProperty],
        potential_difference_based: bool,
    ):
        """
        Constructor.

        :param name: the uniquely identifying name of this component
        :param prop: the BoundedProperty for which a value will be retrieved
            from the State
        :param state_variables: the state variables corresponding to each State element
            that this component will be passed.
        :param is_potential_based: True if reward is based on a potential difference
            between prev_state and state (AKA potential based shaping reward) else
            False (and reward depends only on the potential of current state).
        """
        self.name = name
        self.state_index_of_value = state_variables.index(prop)
        self.potential_difference_based = potential_difference_based
        self._set_target(target, state_variables)

    def _set_target(
        self,
        target: Union[int, float, prp.Property, prp.BoundedProperty],
        state_variables: Tuple[prp.BoundedProperty],
    ) -> None:
        """
        Sets the target value or an index for retrieving it from States

        Depending on how target is specified, it may either be a constant, or a
        Property's value that needs to be retrieved from the State.
        """
        if isinstance(target, float) or isinstance(target, int):
            self.constant_target = True
            self.target = target
        elif isinstance(target, prp.Property) or isinstance(
            target, prp.BoundedProperty
        ):
            self.constant_target = False
            self.target_index = state_variables.index(target)

    def calculate(self, state: State, prev_state: State, is_terminal: bool):
        """
        Calculates the value of this RewardComponent.

        If this component is potential difference based, its value is the
        difference in potentials between prev_state and state. Otherwise its
        value is the potential of state.
        """
        if self.potential_difference_based:
            # reward is a potential difference of state, prev_state
            reward = self.get_potential(state, is_terminal) - self.get_potential(
                prev_state, False
            )
        else:
            reward = self.get_potential(state, is_terminal)
        return reward

    def is_constant_target(self):
        return self.constant_target

    def get_name(self) -> str:
        return self.name

    def is_potential_difference_based(self) -> bool:
        return self.potential_difference_based


class ErrorComponent(NormalisedComponent, ABC):
    """
    Calculates rewards based on a normalised error complement from a target value.

    Normalising an error takes some absolute difference |value - target| and
    transforms it to the interval [0,1], where 1.0 is no error and 0.0 is inf error.
    """

    def get_potential(self, state: State, is_terminal) -> float:
        """
        Calculates the 'goodness' of a State given we want the compare_property
        to be some target_value. The target value may be a constant or
        retrieved from another property in the state.

        The 'goodness' of the state is given in the interval [-1,0], where 0
        corresponds to zero error, and -1 corresponds to inf error.
        """
        if is_terminal and self.potential_difference_based:
            return self.POTENTIAL_BASED_DIFFERENCE_TERMINAL_VALUE

        if self.is_constant_target():
            target = self.target
        else:
            # else we have to look it up from the state
            target = state[self.target_index]
        value = state[self.state_index_of_value]
        error = abs(value - target)
        return 1 - self._normalise_error(error)

    @abstractmethod
    def _normalise_error(self, absolute_error: float) -> float:
        """
        Given an error in the interval [0, +inf], returns a normalised error in [0, 1]

        The normalised error asymptotically approaches 1 as absolute_error -> +inf.

        The parameter error_scaling is used to scale for magnitude.
        When absolute_error == error_scaling, the normalised error is equal to 0.5
        """
        ...


class AsymptoticErrorComponent(ErrorComponent):
    """
    A reward component which gives a negative reward that asymptotically approaches -1
    as the error to the desired value approaches +inf. This is convenient for not having
    to worry about the bounds on the absolute error value.
    """

    def __init__(
        self,
        name: str,
        prop: prp.BoundedProperty,
        state_variables: Tuple[prp.BoundedProperty],
        target: Union[int, float, prp.Property, prp.BoundedProperty],
        is_potential_based: bool,
        scaling_factor: Union[float, int],
    ):
        """
        Constructor.

        :param scaling_factor: the property value is scaled down by this amount.
            Shaping potential is at 0.5 when the error equals this factor.
        """
        super().__init__(name, prop, state_variables, target, is_potential_based)
        self.scaling_factor = scaling_factor

    def _normalise_error(self, absolute_error: float):
        return normalise_error_asymptotic(absolute_error, self.scaling_factor)


class AngularAsymptoticErrorComponent(AsymptoticErrorComponent):
    """
    A potential-based shaping reward component.

    Potential is based asymptotically on the  size of the error between a
    property of interest and its target. The error can be unbounded in
    magnitude.

    Values must be in units of degrees. Errors are reduced to the interval
    (-180, 180] before processing.
    """

    def _normalise_error(self, angular_error: float):
        """
        Given an angle off of a target direction in degrees, calculates a
        normalised error in [0,1]. The angular error is firstly transformed
        to interval [-180,180] to account for the fact the agent can turn
        left or right to face the target.

        :param angular_error: float, angle off target in degrees
        :return: float, normalised error in [0,1]
        """
        reduced_angle_error = abs(reduce_reflex_angle_deg(angular_error))
        return super()._normalise_error(reduced_angle_error)


class LinearErrorComponent(ErrorComponent):
    """
    A potential-based shaping reward component.

    Potential is based linearly on the size of the error between a property of
    interest and its target. The error must be in the interval [0, scaling_factor].
    """

    def __init__(
        self,
        name: str,
        prop: prp.BoundedProperty,
        state_variables: Tuple[prp.BoundedProperty],
        target: Union[int, float, prp.Property, prp.BoundedProperty],
        is_potential_based: bool,
        scaling_factor: Union[float, int],
    ):
        """
        Constructor.

        :param scaling_factor: the max size of the difference between prop and
            target. Minimum potential (0.0) occurs when error is
            max_error_size or greater.
        """
        super().__init__(name, prop, state_variables, target, is_potential_based)
        self.scaling_factor = scaling_factor

    def _normalise_error(self, absolute_error: float):
        return normalise_error_linear(absolute_error, self.scaling_factor)


def normalise_error_asymptotic(absolute_error: float, scaling_factor: float) -> float:
    """
    Given an error in the interval [0, +inf], returns a normalised error in [0, 1]

    The normalised error asymptotically approaches 1 as absolute_error -> +inf.

    The parameter scaling_factor is used to scale for magnitude.
    When absolute_error == scaling_factor, the normalised error is equal to 0.5
    """
    if absolute_error < 0:
        raise ValueError(
            f"Error to be normalised must be non-negative : {absolute_error}"
        )
    scaled_error = absolute_error / scaling_factor
    return scaled_error / (scaled_error + 1)


def normalise_error_linear(absolute_error: float, max_error: float) -> float:
    """
    Given an absolute error in [0, max_error], linearly normalises error in [0, 1]

    If absolute_error exceeds max_error, it is capped back to max_error
    """
    if absolute_error < 0:
        raise ValueError(
            f"Error to be normalised must be non-negative : {absolute_error}"
        )
    elif absolute_error > max_error:
        return 1.0
    else:
        return absolute_error / max_error


class RewardStub(Reward):
    def __init__(self, agent_reward_value: float, assessment_reward_value: float):
        assert isinstance(agent_reward_value, float)
        assert isinstance(assessment_reward_value, float)
        self.agent_reward_value = agent_reward_value
        self.assessment_reward_value = assessment_reward_value

    def agent_reward(self) -> float:
        return self.agent_reward_value

    def assessment_reward(self) -> float:
        return self.assessment_reward_value

    def is_shaping(self):
        return True



import gymnasium as gym
import numpy as np
from jsbgym.tasks import Shaping, HeadingControlTask
from jsbgym.simulation import Simulation
from jsbgym.visualiser import FigureVisualiser, FlightGearVisualiser
from jsbgym.aircraft import Aircraft, cessna172P
from typing import Optional, Type, Tuple, Dict


class JsbSimEnv(gym.Env):
    """
    A class wrapping the JSBSim flight dynamics module (FDM) for simulating
    aircraft as an RL environment conforming to the Gymnasium Env
    interface.

    An JsbSimEnv is instantiated with a Task that implements a specific
    aircraft control task with its own specific observation/action space and
    variables and agent_reward calculation.

    ATTRIBUTION: this class implements the Gymnasium Env API. Method
    docstrings have been adapted or copied from the OpenAI Gym source code then converted to work with the gymnasium interface.
    """

    JSBSIM_DT_HZ: int = 60  # JSBSim integration frequency
    metadata = {
        "render_modes": ["human", "flightgear", "human_fg", "rgb_array"],
        "render_fps": 60,
    }

    def __init__(
        self,
        aircraft: Aircraft = cessna172P,
        task_type: Type = HeadingControlTask,
        agent_interaction_freq: int = 5,
        shaping: Shaping = Shaping.STANDARD,
        render_mode: Optional[str] = None,
    ):
        """
        Constructor. Inits some internal state, but JsbSimEnv.reset() must be
        called first before interacting with environment.

        :param task_type: the Task subclass for the task agent is to perform
        :param aircraft: the JSBSim aircraft to be used
        :param agent_interaction_freq: int, how many times per second the agent
            should interact with environment.
        :param shaping: a HeadingControlTask.Shaping enum, what type of agent_reward
            shaping to use (see HeadingControlTask for options)
        """
        if agent_interaction_freq > self.JSBSIM_DT_HZ:
            raise ValueError(
                "agent interaction frequency must be less than "
                "or equal to JSBSim integration frequency of "
                f"{self.JSBSIM_DT_HZ} Hz."
            )
        self.sim: Simulation = None
        self.sim_steps_per_agent_step: int = self.JSBSIM_DT_HZ // agent_interaction_freq
        self.aircraft = aircraft
        self.task = task_type(shaping, agent_interaction_freq, aircraft)
        # set Space objects
        self.observation_space: gym.spaces.Box = self.task.get_state_space()
        self.action_space: gym.spaces.Box = self.task.get_action_space()
        # set visualisation objects
        self.figure_visualiser: FigureVisualiser = None
        self.flightgear_visualiser: FlightGearVisualiser = None
        self.step_delay = None
        self.render_mode = render_mode

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, Dict]:
        if self.render_mode == "human":
            self.render()
        """
        Run one timestep of the environment's dynamics. When end of
        episode is reached, you are responsible for calling `reset()`
        to reset this environment's state.
        Accepts an action and returns a tuple (observation, reward, terminated, False, info).

        :param action: the agent's action, with same length as action variables.
        :return:
            state: agent's observation of the current environment
            reward: amount of reward returned after previous action
            terminated: whether the episode has ended, in which case further step() calls are undefined
            False: Truncated
            info: auxiliary information, e.g. full reward shaping data
        """
        if action.shape != self.action_space.shape:
            raise ValueError("mismatch between action and action space size")

        state, reward, terminated, truncated, info = self.task.task_step(
            self.sim, action, self.sim_steps_per_agent_step
        )
        observation = np.array(state)
        return observation, reward, terminated, False, info

    def reset(self, seed: Optional[int] = None, options: Optional[dict] = None):
        """
        Resets the state of the environment and returns an initial observation.

        :return: array, the initial observation of the space.
        """
        super().reset(seed=seed)
        init_conditions = self.task.get_initial_conditions()
        if self.sim:
            self.sim.reinitialise(init_conditions)
        else:
            self.sim = self._init_new_sim(
                self.JSBSIM_DT_HZ, self.aircraft, init_conditions
            )

        state = self.task.observe_first_state(self.sim)

        if self.flightgear_visualiser:
            self.flightgear_visualiser.configure_simulation_output(self.sim)
        observation = np.array(state)
        info = {}
        if self.render_mode == "human":
            self.render()
        return observation, info

    def _init_new_sim(self, dt, aircraft, initial_conditions):
        return Simulation(
            sim_frequency_hz=dt, aircraft=aircraft, init_conditions=initial_conditions
        )

    def render(self, flightgear_blocking=True):
        if self.render_mode is None:
            assert self.spec is not None
            gym.logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. gym.make("{self.spec.id}", render_mode="human")'
            )
            return

        """Renders the environment.
        The set of supported modes varies per environment. (And some
        environments do not support rendering at all.) By convention,
        if mode is:
        - human: render to the current display or terminal and
          return nothing. Usually for human consumption.
        - rgb_array: Return an numpy.ndarray with shape (x, y, 3),
          representing RGB values for an x-by-y pixel image, suitable
          for turning into a video.
        - ansi: Return a string (str) or StringIO.StringIO containing a
          terminal-style text representation. The text can include newlines
          and ANSI escape sequences (e.g. for colors).
        Note:
            Make sure that your class's metadata 'render.modes' key includes
              the list of supported modes. It's recommended to call super()
              in implementations to use the functionality of this method.

        :param mode: str, the mode to render with
        :param flightgear_blocking: waits for FlightGear to load before
            returning if True, else returns immediately
        """

        if self.render_mode == "human":
            if not self.figure_visualiser:
                self.figure_visualiser = FigureVisualiser(
                    self.sim, self.task.get_props_to_output()
                )
            self.figure_visualiser.plot(self.sim)
        elif self.render_mode == "flightgear":
            if not self.flightgear_visualiser:
                self.flightgear_visualiser = FlightGearVisualiser(
                    self.sim, self.task.get_props_to_output(), flightgear_blocking
                )
        elif self.render_mode == "human_fg":
            if not self.flightgear_visualiser:
                self.flightgear_visualiser = FlightGearVisualiser(
                    self.sim, self.task.get_props_to_output(), flightgear_blocking
                )
            self.flightgear_visualiser.plot(self.sim)
        else:
            super().render()

    def close(self):
        """Cleans up this environment's objects

        Environments automatically close() when garbage collected or when the
        program exits.
        """
        if self.sim:
            self.sim.close()
        if self.figure_visualiser:
            self.figure_visualiser.close()
        if self.flightgear_visualiser:
            self.flightgear_visualiser.close()


class NoFGJsbSimEnv(JsbSimEnv):
    """
    An RL environment for JSBSim with rendering to FlightGear disabled.
    This class exists to be used for training agents where visualisation is not
    required. Otherwise, restrictions in JSBSim output initialisation cause it
    to open a new socket for every single episode, eventually leading to
    failure of the network.
    """

    JSBSIM_DT_HZ: int = 60  # JSBSim integration frequency
    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 60,
    }

    def _init_new_sim(self, dt: float, aircraft: Aircraft, initial_conditions: Dict):
        return Simulation(
            sim_frequency_hz=dt,
            aircraft=aircraft,
            init_conditions=initial_conditions,
            allow_flightgear_output=False,
        )

    def render(self, flightgear_blocking=True):
        if self.render_mode == "flightgear" or self.render_mode == "human_fg":
            raise ValueError("FlightGear rendering is disabled for this class")
        else:
            super().render(flightgear_blocking)



import warnings
from jsbgym import utils
from abc import ABC, abstractmethod
from typing import Iterable, Tuple, Dict
from jsbgym.rewards import State, Reward, RewardComponent


class Assessor(ABC):
    """Interface for Assessors which calculate Rewards from States."""

    @abstractmethod
    def assess(self, state: State, prev_state: State, is_terminal: bool) -> Reward:
        """Calculates reward from environment's state, previous state and terminal condition"""
        ...


class AssessorImpl(Assessor):
    """
    Determines the Reward from a state transitions.

    Initialised with RewardComponents which allow calculation of the base
    (policy-influencing) and shaping rewards (non-policy-influencing) rewards respectively.
    """

    def __init__(
        self,
        base_components: Iterable["RewardComponent"],
        potential_based_components: Iterable["RewardComponent"] = (),
        positive_rewards: bool = False,
    ):
        """
        :param base_components: RewardComponents from which Reward is to be calculated
        :param potential_based_components: RewardComponents from which a potential-based
            reward component is to be calculated from
        :param positive_rewards: True if rewards should be in [0.0, 1.0] (0.0 corresp. to
            worst behaviour), else rewards will be in [-1.0, 0.0] with 0.0 corresp. to
            perfect behaviour. Has no effect one potential difference based components.
        """
        self.base_components = tuple(base_components)
        self.potential_components = tuple(potential_based_components)
        self.positive_rewards = positive_rewards
        if not self.base_components:
            raise ValueError("base reward components cannot be empty")
        if any(cmp.is_potential_difference_based() for cmp in self.base_components):
            raise ValueError(
                "base rewards must be non potential based in this implementation"
            )
            # because of the positive_rewards logic
        if not all(
            cmp.is_potential_difference_based() for cmp in self.potential_components
        ):
            warnings.warn(f"Potential component not is_potential_difference_based()")

    def assess(self, state: State, prev_state: State, is_terminal: bool) -> Reward:
        """Calculates a Reward from the state transition."""
        return Reward(
            self._base_rewards(state, prev_state, is_terminal),
            self._potential_based_rewards(state, prev_state, is_terminal),
        )

    def _base_rewards(
        self, state: State, prev_state: State, is_terminal: bool
    ) -> Tuple[float, ...]:
        cmp_values = (
            cmp.calculate(state, prev_state, is_terminal)
            for cmp in self.base_components
        )
        if self.positive_rewards:
            return tuple(cmp_values)
        else:
            return tuple(value - 1 for value in cmp_values)

    def _potential_based_rewards(
        self, state: State, last_state: State, is_terminal: bool
    ) -> Tuple[float, ...]:
        return tuple(
            cmp.calculate(state, last_state, is_terminal)
            for cmp in self.potential_components
        )


class SequentialAssessor(AssessorImpl, ABC):
    """
    Abstract class that allows base and potential components to be assigned
    dependencies of other components, such that they are affected by the
    other's values.

    Concrete subclasses should implement _apply_dependents(), which modifies
    the 'normal' component potentials to account for dependents
    """

    def __init__(
        self,
        base_components: Iterable["RewardComponent"],
        potential_components: Iterable["RewardComponent"] = (),
        base_dependency_map: Dict[
            "RewardComponent", Tuple["RewardComponent", ...]
        ] = {},
        potential_dependency_map: Dict[
            "RewardComponent", Tuple["RewardComponent", ...]
        ] = {},
        positive_rewards: bool = False,
    ):
        """
        :param base_components: RewardComponents from which the non-shaping
            part of the Reward is to be calculated
        :param potential_components: ErrorComponents from which the shaping
            reward is to be calculated, or an empty tuple for no shaping
        :param base_dependency_map: maps base components with sequential
            dependencies to their dependent components, defaults to
            no dependencies
        :param potential_dependency_map: maps potential components with sequential
            dependencies to their dependent components, defaults to
            no dependencies
        """
        super().__init__(base_components, potential_components, positive_rewards)
        self.base_dependency_map = base_dependency_map
        self.potential_dependency_map = potential_dependency_map

    def _base_rewards(
        self, state: State, prev_state: State, is_terminal: bool
    ) -> Tuple[float, ...]:
        potentials = tuple(
            cmp.get_potential(state, is_terminal) for cmp in self.base_components
        )
        seq_discounts = self._get_sequential_discounts(
            state, is_terminal, self.base_components, self.base_dependency_map
        )

        seq_values = (
            pot * discount for pot, discount in zip(potentials, seq_discounts)
        )
        if self.positive_rewards:
            return tuple(seq_values)
        else:
            return tuple(value - 1 for value in seq_values)

    def _potential_based_rewards(
        self, state: State, prev_state: State, is_terminal: bool
    ) -> Tuple[float, ...]:
        potentials = tuple(
            cmp.get_potential(state, is_terminal) for cmp in self.potential_components
        )
        prev_potentials = tuple(
            cmp.get_potential(prev_state, False) for cmp in self.potential_components
        )

        discounts = self._get_sequential_discounts(
            state, is_terminal, self.potential_components, self.potential_dependency_map
        )
        prev_discounts = self._get_sequential_discounts(
            prev_state, False, self.potential_components, self.potential_dependency_map
        )

        seq_potentials = (p * d for p, d in zip(potentials, discounts))
        seq_prev_potentials = (p * d for p, d in zip(prev_potentials, prev_discounts))
        return tuple(
            pot - prev_pot for pot, prev_pot in zip(seq_potentials, seq_prev_potentials)
        )

    @abstractmethod
    def _get_sequential_discounts(
        self,
        state: State,
        is_terminal: bool,
        components: Iterable["RewardComponent"],
        dependency_map: Dict["RewardComponent", Tuple],
    ) -> Tuple[float, ...]:
        """
        Calculates a discount factor in [0,1] from each component's dependencies.

        The dependencies may reduce that component's values because they are not
        yet met. A component with no dependencies has discount factor 1.0.

        :param state: the state that discount factor is to be evaluated at
        :param is_terminal: whether the transition to state was terminal
        :param components: the RewardComponents to be assessed for discounting
        :param dependency_map: a map of RewardComponents to their dependent
            RewardComponents.
        :return: tuple of floats, discount factors in [0,1], corresponding to
            same order as 'components' input
        """
        ...


class ContinuousSequentialAssessor(SequentialAssessor):
    """
    A sequential assessor in which shaping components with dependents have their potential
    reduced according to their dependent's potentials through multiplication.

    For example a component with a "base" potential of 0.8 and a dependent component at
    0.5 have a sequential potential of 0.8 * 0.5 = 0.4.
    """

    def _get_sequential_discounts(
        self,
        state: State,
        is_terminal: bool,
        components: Iterable["RewardComponent"],
        dependency_map: Dict["RewardComponent", Tuple["RewardComponent", ...]],
    ) -> Tuple[float, ...]:
        discounts = []
        for component in components:
            dependents = dependency_map.get(component, ())
            dependent_potentials = (
                dep.get_potential(state, is_terminal) for dep in dependents
            )
            discount = utils.product(pot for pot in dependent_potentials)
            discounts.append(discount)
        return tuple(discounts)



import gymnasium as gym
import enum
from jsbgym.tasks import Task, HeadingControlTask, TurnHeadingControlTask
from jsbgym.aircraft import Aircraft, cessna172P
from jsbgym import utils

"""
This script registers all combinations of task, aircraft, shaping settings
 etc. with OpenAI Gym so that they can be instantiated with a gym.make(id)
 command.

The jsbgym.Envs enum stores all registered environments as members with
 their gym id string as value. This allows convenient autocompletion and value
 safety. To use do:
       env = gym.make(jsbgym.Envs.desired_environment.value)
"""

for env_id, (
    plane,
    task,
    shaping,
    enable_flightgear,
) in utils.get_env_id_kwargs_map().items():
    if enable_flightgear:
        entry_point = "jsbgym.environment:JsbSimEnv"
    else:
        entry_point = "jsbgym.environment:NoFGJsbSimEnv"
    kwargs = dict(aircraft=plane, task_type=task, shaping=shaping)
    gym.envs.registration.register(id=env_id, entry_point=entry_point, kwargs=kwargs)



import unittest
import numpy as np
import gymnasium as gym
from jsbgym import utils
from jsbgym.agents import RandomAgent
from jsbgym.environment import JsbSimEnv
from jsbgym.tasks import HeadingControlTask
import jsbgym.properties as prp


class AgentEnvInteractionTest(unittest.TestCase):
    """Tests for agents interacting with env."""

    def init_and_reset_env(self, env: JsbSimEnv):
        self.assertIsInstance(env.task, HeadingControlTask)

        # we interact at 5 Hz, so we expect the sim to run 12 timesteps per
        #   interaction since it runs at 120 Hz
        self.assertEqual(12, env.sim_steps_per_agent_step)

        # we init a random agent with a seed
        agent = RandomAgent(action_space=env.action_space)
        self.assertEqual(env.action_space, agent.action_space)

        # this task has an action space of three controls: aileron, elevator, rudder
        expected_num_actions = 3
        self.assertEqual(expected_num_actions, len(agent.action_space.low))
        # we see that the action space has the correct low and high range of +-1.0
        expect_low = np.array([-1.0] * expected_num_actions)
        expect_high = np.array([1.0] * expected_num_actions)
        np.testing.assert_array_almost_equal(expect_high, env.action_space.high)
        np.testing.assert_array_almost_equal(expect_low, env.action_space.low)

        # we reset the env and receive the first state; the env is now ready
        state = env.reset()
        self.assertEqual(len(env.observation_space.low), len(state))

        # we close the env and JSBSim closes with it
        env.close()
        self.assertIsNone(env.sim.jsbsim)

    def take_step_with_random_agent(self, env: JsbSimEnv):
        agent = RandomAgent(action_space=env.action_space)

        # we set up for a loop through one episode
        first_state = env.reset()

        # we take a single step
        action = agent.act(first_state)
        state, reward, terminated, truncated, info = env.step(action)

        # we see the state has changed
        self.assertEqual(first_state.shape, state.shape)
        self.assertTrue(
            np.any(np.not_equal(first_state, state)),
            msg="state should have changed after simulation step",
        )
        expected_time_step_size = env.sim_steps_per_agent_step / env.JSBSIM_DT_HZ
        self.assertAlmostEqual(expected_time_step_size, env.sim.get_sim_time())
        self.assertFalse(terminated, msg="episode is terminal after only a single step")

        # the aircraft engines are running, as per initial conditions
        self.assertNotAlmostEqual(env.sim[prp.engine_thrust_lbs], 0)

        env.close()

    def test_init_and_reset_all_envs(self):
        for env_id in utils.get_env_id_kwargs_map():
            env = gym.make(env_id)
            self.init_and_reset_env(env)

    def test_take_step_with_random_agent_all_envs(self):
        for env_id in utils.get_env_id_kwargs_map():
            env = gym.make(env_id)
            self.take_step_with_random_agent(env)



import unittest
import math
import numpy as np
import sys
import jsbgym.properties as prp
from jsbgym import rewards, utils
from jsbgym.assessors import Assessor, AssessorImpl
from jsbgym.aircraft import Aircraft, cessna172P
from jsbgym.tasks import Shaping, HeadingControlTask, TurnHeadingControlTask
from jsbgym.tests.stubs import SimStub, TransitioningSimStub


class TestHeadingControlTask(unittest.TestCase):
    default_shaping = Shaping.STANDARD
    default_episode_time_s = 1
    default_step_frequency_hz = 1
    default_aircraft = cessna172P
    default_steps_remaining_non_terminal = 10
    default_positive_rewards = True

    PERFECT_POSITIVE_REWARD = 1.0
    MIDDLING_POSITIVE_REWARD = 0.5
    TERRIBLE_POSITIVE_REWARD = 0.0

    def setUp(self):
        self.task = self.make_task()
        sim = SimStub.make_valid_state_stub(self.task)
        # causes task to init new-episode attributes
        _ = self.task.observe_first_state(sim)

        self.dummy_action = np.asarray(
            [0 for _ in range(len(self.task.action_variables))]
        )

    def get_class_under_test(self):
        return HeadingControlTask

    def make_task(
        self,
        shaping_type: Shaping = default_shaping,
        episode_time_s: float = default_episode_time_s,
        step_frequency_hz: float = default_step_frequency_hz,
        aircraft: Aircraft = default_aircraft,
        positive_rewards: bool = True,
    ) -> HeadingControlTask:
        task_class = self.get_class_under_test()
        return task_class(
            shaping_type=shaping_type,
            episode_time_s=episode_time_s,
            step_frequency_hz=step_frequency_hz,
            aircraft=aircraft,
            positive_rewards=positive_rewards,
        )

    def get_initial_sim_with_state(
        self,
        task: HeadingControlTask = None,
        time_terminal=False,
        track_deg=None,
        altitude_ft=None,
        roll_rad=0.0,
    ) -> SimStub:
        if task is None:
            task = self.task
        sim = SimStub.make_valid_state_stub(task)
        task.observe_first_state(sim)

        if track_deg is None:
            # get the target value and use that (i.e. perfect state)
            track_deg = sim[task.target_track_deg]

        perfect_initial_sim = self.modify_sim_to_state_(
            sim, task, time_terminal, track_deg, altitude_ft, roll_rad
        )
        # the task set its last_state attr when it observed the first state - reset this
        #   based on the perfect sim values we just set. Ugh.
        perfect_state = task.State(*(sim[prop] for prop in task.state_variables))
        task.last_state = perfect_state
        return perfect_initial_sim

    def modify_sim_to_state_(
        self,
        sim: SimStub,
        task: HeadingControlTask = None,
        steps_terminal=False,
        track_deg=0.0,
        altitude_ft=None,
        roll_rad=0.0,
        sideslip_deg=0.0,
    ) -> SimStub:
        if task is None:
            task = self.task

        if steps_terminal:
            sim[self.task.steps_left] = 0
        else:
            sim[self.task.steps_left] = self.default_steps_remaining_non_terminal

        if altitude_ft is None:
            sim[prp.altitude_sl_ft] = task.INITIAL_ALTITUDE_FT
        else:
            sim[prp.altitude_sl_ft] = altitude_ft

        v_east = math.sin(math.radians(track_deg))
        v_north = math.cos(math.radians(track_deg))
        sim[prp.v_east_fps] = v_east
        sim[prp.v_north_fps] = v_north
        task._update_track_error(sim)
        sim[prp.roll_rad] = roll_rad
        sim[prp.sideslip_deg] = sideslip_deg
        return sim

    def get_initial_state_sim(self, task=None) -> SimStub:
        if task is None:
            task = self.task
        sim = SimStub.make_valid_state_stub(task)

        # set properties to reasonable initial episode values
        sim[prp.sim_time_s] = 0.0
        sim[prp.dist_travel_m] = 0.0
        sim[prp.heading_deg] = task.INITIAL_HEADING_DEG
        sim[prp.altitude_sl_ft] = task.get_initial_conditions()[prp.initial_altitude_ft]
        sim[prp.roll_rad] = 0
        return sim

    def get_perfect_state_sim(self, task=None, time_terminal=True) -> SimStub:
        if task is None:
            task = self.task
        sim = SimStub.make_valid_state_stub(task)
        task.observe_first_state(sim)

        perfect_track = sim[task.target_track_deg]
        perfect_altitude = task._get_target_altitude()
        perfect_roll = 0.0
        perfect_sideslip = 0.0

        self.modify_sim_to_state_(
            sim,
            task,
            steps_terminal=time_terminal,
            track_deg=perfect_track,
            altitude_ft=perfect_altitude,
            roll_rad=perfect_roll,
            sideslip_deg=perfect_sideslip,
        )

        return sim

    def get_transitioning_sim(
        self, task, steps_terminal, altitude_ft, roll_rad, track_error_deg, sideslip_deg
    ) -> TransitioningSimStub:
        """Makes a sim that transitions between two states.

        The first state is as configured by the input task.
        The second state's altitude, distance traveled, heading etc. changes as per inputs.
        """
        initial_state_sim = self.get_initial_sim_with_state(task, time_terminal=False)
        target_track = initial_state_sim[self.task.target_track_deg]
        track_deg = target_track + track_error_deg
        next_state_sim = initial_state_sim.copy()
        next_state_sim = self.modify_sim_to_state_(
            next_state_sim,
            steps_terminal=steps_terminal,
            altitude_ft=altitude_ft,
            roll_rad=roll_rad,
            track_deg=track_deg,
            sideslip_deg=sideslip_deg,
        )
        return TransitioningSimStub(initial_state_sim, next_state_sim)

    def get_target_track(self, sim: SimStub):
        return self.task.INITIAL_HEADING_DEG

    def test_init_shaping_standard(self):
        task = self.make_task(shaping_type=Shaping.STANDARD)

        self.assertIsInstance(task.assessor, Assessor)
        self.assertEqual(2, len(task.assessor.base_components))
        self.assertEqual(0, len(task.assessor.potential_components))
        self.assertFalse(task.assessor.potential_components)  # assert empty

    def test_init_shaping_sequential(self):
        task = self.make_task(shaping_type=Shaping.EXTRA_SEQUENTIAL)

        self.assertIsInstance(task.assessor, Assessor)
        self.assertEqual(2, len(task.assessor.base_components))
        self.assertEqual(2, len(task.assessor.potential_components))

    def test_get_intial_conditions_valid_target_heading(self):
        self.setUp()

        ics = self.task.get_initial_conditions()
        initial_heading = ics[prp.initial_heading_deg]

        self.assertLessEqual(prp.heading_deg.min, initial_heading)
        self.assertGreaterEqual(prp.heading_deg.max, initial_heading)

    def test_get_initial_conditions_contains_all_props(self):
        ics = self.task.get_initial_conditions()

        self.assertIsInstance(ics, dict)
        for prop, value in self.task.base_initial_conditions.items():
            self.assertAlmostEqual(value, ics[prop])

        steady_level_task_ic_properties = [
            prp.initial_u_fps,
            prp.initial_v_fps,
            prp.initial_w_fps,
            prp.initial_p_radps,
            prp.initial_q_radps,
            prp.initial_r_radps,
            prp.initial_heading_deg,
        ]
        for prop in steady_level_task_ic_properties:
            self.assertIn(
                prop,
                ics.keys(),
                msg="expected HeadingControlTask to set value for"
                f"property {prop} but not found in ICs",
            )

    def test_observe_first_state_returns_valid_state(self):
        sim = SimStub.make_valid_state_stub(self.task)

        first_state = self.task.observe_first_state(sim)

        self.assertEqual(len(first_state), len(self.task.state_variables))
        self.assertIsInstance(first_state, tuple)

    def test_task_first_observation_inputs_controls(self):
        dummy_sim = SimStub.make_valid_state_stub(self.task)
        _ = self.task.observe_first_state(dummy_sim)

        # check engine as expected
        self.assertAlmostEqual(self.task.THROTTLE_CMD, dummy_sim[prp.throttle_cmd])
        self.assertAlmostEqual(self.task.MIXTURE_CMD, dummy_sim[prp.mixture_cmd])
        self.assertAlmostEqual(1.0, dummy_sim[prp.engine_running])

    def test_task_step_correct_return_types(self):
        sim = SimStub.make_valid_state_stub(self.task)
        steps = 1
        _ = self.task.observe_first_state(sim)

        state, reward, is_terminal, truncated, info = self.task.task_step(
            sim, self.dummy_action, steps
        )

        self.assertIsInstance(state, tuple)
        self.assertEqual(len(state), len(self.task.state_variables))

        self.assertIsInstance(reward, float)
        self.assertIsInstance(is_terminal, bool)
        self.assertIsInstance(info, dict)

    def test_task_step_returns_reward_in_info(self):
        sim = SimStub.make_valid_state_stub(self.task)
        steps = 1
        _ = self.task.observe_first_state(sim)

        _, reward_scalar, _, info = self.task.task_step(sim, self.dummy_action, steps)
        reward_object = info["reward"]

        self.assertIsInstance(reward_object, rewards.Reward)
        self.assertAlmostEqual(reward_object.agent_reward(), reward_scalar)

    def test_task_step_returns_non_terminal_time_less_than_max(self):
        sim = self.get_perfect_state_sim(self.task, time_terminal=False)
        _ = self.task.observe_first_state(sim)
        non_terminal_steps_left = 2
        sim[self.task.steps_left] = non_terminal_steps_left

        _, _, is_terminal, _ = self.task.task_step(sim, self.dummy_action, 1)

        self.assertFalse(is_terminal)

    def test_task_step_returns_terminal_time_exceeds_max(self):
        sim = SimStub.make_valid_state_stub(self.task)
        _ = self.task.observe_first_state(sim)
        terminal_time = self.default_episode_time_s + 1
        sim[prp.sim_time_s] = terminal_time
        steps = 1

        _, _, is_terminal, _ = self.task.task_step(sim, self.dummy_action, steps)

        self.assertTrue(is_terminal)

    def test_task_step_returns_terminal_time_equals_max(self):
        sim = SimStub.make_valid_state_stub(self.task)
        _ = self.task.observe_first_state(sim)
        terminal_time = self.default_episode_time_s
        sim[prp.sim_time_s] = terminal_time
        steps = 1

        _, _, is_terminal, _ = self.task.task_step(sim, self.dummy_action, steps)

        self.assertTrue(is_terminal)

    def test_task_step_correct_terminal_reward_optimal_behaviour_no_shaping(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.STANDARD,
                episode_time_s=1.0,
                step_frequency_hz=1.0,
                positive_rewards=positive_reward,
            )
            initial_state_sim = self.get_initial_state_sim(task)
            _ = task.observe_first_state(initial_state_sim)
            final_state_sim = self.get_perfect_state_sim(task, time_terminal=True)
            sim = TransitioningSimStub(initial_state_sim, final_state_sim)

            state, reward, terminated, truncated, info = task.task_step(
                sim, self.dummy_action, 1
            )

            if positive_reward:
                expected_reward = self.PERFECT_POSITIVE_REWARD
            else:
                expected_reward = self.PERFECT_POSITIVE_REWARD - 1

            self.assertAlmostEqual(expected_reward, reward)

    def test_task_step_correct_non_terminal_reward_optimal_behaviour_no_shaping(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.STANDARD, positive_rewards=positive_reward
            )
            initial_state_sim = self.get_initial_state_sim(task)
            _ = task.observe_first_state(initial_state_sim)
            final_state_sim = self.get_perfect_state_sim(task, time_terminal=False)
            sim = TransitioningSimStub(initial_state_sim, final_state_sim)

            state, reward, terminated, truncated, info = task.task_step(
                sim, self.dummy_action, 1
            )

            if positive_reward:
                expected_reward = self.PERFECT_POSITIVE_REWARD
            else:
                expected_reward = self.PERFECT_POSITIVE_REWARD - 1
            self.assertAlmostEqual(expected_reward, reward)

    def test_task_step_correct_terminal_reward_optimal_behaviour_shaping(self):
        for positive_reward in (True, False):
            self.setUp()
            for shaping in (Shaping.STANDARD, Shaping.EXTRA_SEQUENTIAL):
                task = self.make_task(
                    shaping_type=shaping, positive_rewards=positive_reward
                )
                initial_state_sim = self.get_initial_state_sim(task)
                _ = task.observe_first_state(initial_state_sim)
                final_state_sim = self.get_perfect_state_sim(task, time_terminal=True)
                sim = TransitioningSimStub(initial_state_sim, final_state_sim)

                _, _, _, info = task.task_step(sim, self.dummy_action, 1)
                reward_obj: rewards.Reward = info["reward"]

                if positive_reward:
                    expected_reward = self.PERFECT_POSITIVE_REWARD
                else:
                    expected_reward = self.PERFECT_POSITIVE_REWARD - 1
                self.assertAlmostEqual(expected_reward, reward_obj.assessment_reward())

    def test_task_step_out_of_bounds_altitude_reward_override(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.STANDARD, positive_rewards=positive_reward
            )
            bad_altitude = sys.float_info.max
            sim = self.get_transitioning_sim(
                task,
                steps_terminal=False,
                altitude_ft=bad_altitude,
                roll_rad=0.0,
                track_error_deg=0.0,
                sideslip_deg=0.0,
            )

            _, reward_scalar, _, info = task.task_step(sim, self.dummy_action, 1)
            reward_obj: rewards.Reward = info["reward"]

            if positive_reward:
                # no reward override when using positive reward; perfect track and terrible alt
                expected_reward = (
                    self.PERFECT_POSITIVE_REWARD + self.TERRIBLE_POSITIVE_REWARD
                ) / 2
            else:
                # big negative reward for out of bounds
                expected_reward = -1 + -1.0 * sim[self.task.steps_left]
            self.assertAlmostEqual(expected_reward, reward_scalar)
            self.assertAlmostEqual(expected_reward, reward_obj.agent_reward())

    def test_task_step_middling_track_otherwise_perfect(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.STANDARD, positive_rewards=positive_reward
            )
            perfect_altitude = task._get_target_altitude()
            middling_track = HeadingControlTask.TRACK_ERROR_SCALING_DEG
            sim = self.get_transitioning_sim(
                task,
                steps_terminal=True,
                altitude_ft=perfect_altitude,
                roll_rad=0.0,
                track_error_deg=middling_track,
                sideslip_deg=0.0,
            )

            _, _, _, info = task.task_step(sim, self.dummy_action, 1)
            reward_obj: rewards.Reward = info["reward"]

            expected_positive_reward = (
                self.MIDDLING_POSITIVE_REWARD + self.PERFECT_POSITIVE_REWARD
            ) / 2
            if positive_reward:
                expected_reward = expected_positive_reward
            else:
                expected_reward = expected_positive_reward - 1
            self.assertAlmostEqual(expected_reward, reward_obj.agent_reward())

    def test_task_step_reward_middling_everything(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.STANDARD, positive_rewards=positive_reward
            )
            # we get 0.5 reward at scaling distance from target altitude
            middling_altitude = task.INITIAL_ALTITUDE_FT + task.ALTITUDE_SCALING_FT
            middling_track = HeadingControlTask.TRACK_ERROR_SCALING_DEG
            middling_roll = HeadingControlTask.ROLL_ERROR_SCALING_RAD
            middling_sideslip = HeadingControlTask.SIDESLIP_ERROR_SCALING_DEG
            sim = self.get_transitioning_sim(
                task,
                steps_terminal=True,
                altitude_ft=middling_altitude,
                roll_rad=middling_roll,
                track_error_deg=middling_track,
                sideslip_deg=middling_sideslip,
            )

            _, _, _, info = task.task_step(sim, self.dummy_action, 1)
            reward_obj: rewards.Reward = info["reward"]

            if positive_reward:
                middling_reward = self.MIDDLING_POSITIVE_REWARD
            else:
                middling_reward = self.MIDDLING_POSITIVE_REWARD - 1
            expected_reward = (middling_reward * 4) / 4

            [
                self.assertAlmostEqual(middling_reward, comp)
                for comp in reward_obj.base_reward_elements
            ]
            self.assertAlmostEqual(expected_reward, reward_obj.agent_reward())

    def test_observe_first_state_correct_track_error(self):
        self.setUp()
        sim = SimStub.make_valid_state_stub(self.task)
        _ = self.task.observe_first_state(sim)

        track_deg = prp.Vector2.from_sim(sim).heading_deg()
        target_track_deg = sim[self.task.target_track_deg]
        error_deg = track_deg - target_track_deg
        expected_acute_error_deg = utils.reduce_reflex_angle_deg(error_deg)
        actual_error_deg = sim[self.task.track_error_deg]

        self.assertAlmostEqual(expected_acute_error_deg, actual_error_deg)

    def test_task_step_reward_sequential_cont_perfect(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.EXTRA_SEQUENTIAL, positive_rewards=positive_reward
            )
            perfect_alt = task.INITIAL_ALTITUDE_FT
            perfect_roll, perfect_sideslip, perfect_track_error = 0.0, 0.0, 0.0
            sim = self.get_transitioning_sim(
                task,
                steps_terminal=False,
                altitude_ft=perfect_alt,
                roll_rad=perfect_roll,
                track_error_deg=perfect_track_error,
                sideslip_deg=perfect_sideslip,
            )

            _, reward_scalar, _, info = task.task_step(sim, self.dummy_action, 1)
            reward_obj: rewards.Reward = info["reward"]

            # no improvement (initial state was perfect)
            expected_shaping_reward = 0
            if positive_reward:
                expected_base_reward = self.PERFECT_POSITIVE_REWARD
                expected_reward = (expected_base_reward + expected_shaping_reward) / 2
            else:
                expected_base_reward = self.PERFECT_POSITIVE_REWARD - 1
                expected_reward = (expected_base_reward + expected_shaping_reward) / 2
            msg = f"positive reward: {positive_reward}"
            self.assertAlmostEqual(expected_reward, reward_scalar, msg=msg)
            self.assertAlmostEqual(expected_reward, reward_obj.agent_reward(), msg=msg)
            self.assertAlmostEqual(expected_base_reward, reward_obj.assessment_reward())

    def test_task_step_reward_sequential_extra_shaping_all_middling(self):
        for positive_reward in (True, False):
            self.setUp()
            task = self.make_task(
                shaping_type=Shaping.EXTRA_SEQUENTIAL, positive_rewards=positive_reward
            )
            middling_alt = task.INITIAL_ALTITUDE_FT + task.ALTITUDE_SCALING_FT
            middling_roll = task.ROLL_ERROR_SCALING_RAD
            middling_sideslip = task.SIDESLIP_ERROR_SCALING_DEG
            middling_track_error = task.TRACK_ERROR_SCALING_DEG
            sim = self.get_transitioning_sim(
                task,
                steps_terminal=False,
                altitude_ft=middling_alt,
                roll_rad=middling_roll,
                track_error_deg=middling_track_error,
                sideslip_deg=middling_sideslip,
            )

            _, reward_scalar, _, info = task.task_step(sim, self.dummy_action, 1)
            reward_obj: rewards.Reward = info["reward"]

            # 2 base reward components are both middling
            expected_base_reward = (self.MIDDLING_POSITIVE_REWARD * 2) / 2
            # shaping rewards went from perfect to 1 middling with no deps (.5),
            #   1 middling with a middling dep (.5 * .5)
            expected_shaping_reward = ((0.5 - 1) + (0.5**2 - 1.0)) / 2
            if positive_reward:
                expected_reward = (expected_base_reward + expected_shaping_reward) / 2
            else:
                expected_base_reward = expected_base_reward - 1
                expected_reward = (expected_base_reward + expected_shaping_reward) / 2
            msg = f"positive reward: {positive_reward}"
            self.assertAlmostEqual(expected_reward, reward_scalar, msg=msg)
            self.assertAlmostEqual(expected_reward, reward_obj.agent_reward(), msg=msg)
            self.assertAlmostEqual(expected_base_reward, reward_obj.assessment_reward())

    def test_task_step_reward_on_terminal_shaping(self):
        for positive_reward in (True, False):
            for shaping in (Shaping.EXTRA, Shaping.EXTRA_SEQUENTIAL):
                self.setUp()
                task = self.make_task(
                    shaping_type=shaping, positive_rewards=positive_reward
                )
                perfect_alt = task.INITIAL_ALTITUDE_FT
                perfect_roll, perfect_sideslip, perfect_track_error = 0.0, 0.0, 0.0
                sim = self.get_transitioning_sim(
                    task,
                    steps_terminal=True,
                    altitude_ft=perfect_alt,
                    roll_rad=perfect_roll,
                    track_error_deg=perfect_track_error,
                    sideslip_deg=perfect_sideslip,
                )

                _, reward_scalar, _, info = task.task_step(sim, self.dummy_action, 1)
                reward_obj: rewards.Reward = info["reward"]

                # went from perfect (1.) to terminal (0.)
                expected_shaping_reward = -1
                if positive_reward:
                    expected_base_reward = self.PERFECT_POSITIVE_REWARD
                    expected_reward = (
                        expected_base_reward + expected_shaping_reward
                    ) / 2
                else:
                    expected_base_reward = self.PERFECT_POSITIVE_REWARD - 1
                    expected_reward = (
                        expected_base_reward + expected_shaping_reward
                    ) / 2
                msg = f"positive reward: {positive_reward}"
                self.assertAlmostEqual(expected_reward, reward_scalar, msg=msg)
                self.assertAlmostEqual(
                    expected_reward, reward_obj.agent_reward(), msg=msg
                )
                self.assertAlmostEqual(
                    expected_base_reward, reward_obj.assessment_reward()
                )


class TestTurnHeadingControlTask(TestHeadingControlTask):
    def setUp(self):
        super().setUp()
        assert isinstance(self.task, TurnHeadingControlTask)

    def get_class_under_test(self):
        return TurnHeadingControlTask

    def get_target_track(self, sim: SimStub):
        return sim[TurnHeadingControlTask.target_track_deg]

    def test_observe_first_state_creates_target_heading_in_expected_range(self):
        sim = SimStub.make_valid_state_stub(self.task)
        _ = self.task.observe_first_state(sim)

        desired_heading = sim[HeadingControlTask.target_track_deg]
        self.assertGreaterEqual(desired_heading, 0)
        self.assertLessEqual(desired_heading, 360)

    def test_observe_first_state_changes_target_heading(self):
        sim = SimStub.make_valid_state_stub(self.task)
        _ = self.task.observe_first_state(sim)
        desired_heading = sim[HeadingControlTask.target_track_deg]

        _ = self.task.observe_first_state(sim)
        new_desired_heading = sim[HeadingControlTask.target_track_deg]

        self.assertNotEqual(desired_heading, new_desired_heading)



from jsbgym.properties import GeodeticPosition
import unittest


class TestGeodeticPosition(unittest.TestCase):
    BATH_LAT = 51.3751
    BATH_LNG = -2.36172
    # we have to accept small loss of accuracy (max +-1 degree) due to using Cartesian projection
    heading_accuracy_places = 0

    def setUp(self, lat=BATH_LAT, lng=BATH_LNG):
        self.position = GeodeticPosition(lat, lng)

    def test_heading_deg_to_north_and_south(self):
        lat, lng = self.BATH_LAT, self.BATH_LNG
        north_lat = lat + 1
        south_position = GeodeticPosition(lat, lng)
        north_position = GeodeticPosition(north_lat, lng)

        north_heading_deg = south_position.heading_deg_to(north_position)
        south_heading_deg = north_position.heading_deg_to(south_position)

        self.assertAlmostEqual(
            0, north_heading_deg, places=self.heading_accuracy_places
        )
        self.assertAlmostEqual(
            180, south_heading_deg, places=self.heading_accuracy_places
        )

    def test_heading_deg_to_east_and_west(self):
        lat, lng = self.BATH_LAT, self.BATH_LNG
        east_lng = lng + 1
        west_position = GeodeticPosition(lat, lng)
        east_position = GeodeticPosition(lat, east_lng)

        east_heading_deg = west_position.heading_deg_to(east_position)
        west_heading_deg = east_position.heading_deg_to(west_position)

        self.assertAlmostEqual(
            90, east_heading_deg, places=self.heading_accuracy_places
        )
        self.assertAlmostEqual(
            270, west_heading_deg, places=self.heading_accuracy_places
        )

    def test_heading_deg_ne_and_sw(self):
        lat, lng = self.BATH_LAT, self.BATH_LNG
        north_lat, east_lng = lat + 1, lng + 1
        north_east_position = GeodeticPosition(north_lat, east_lng)
        south_west_position = GeodeticPosition(lat, lng)

        north_east_heading_deg = south_west_position.heading_deg_to(north_east_position)
        south_west_heading_deg = north_east_position.heading_deg_to(south_west_position)

        self.assertAlmostEqual(
            45, north_east_heading_deg, places=self.heading_accuracy_places
        )
        self.assertAlmostEqual(
            225, south_west_heading_deg, places=self.heading_accuracy_places
        )



import unittest
import collections
from typing import Type, NamedTuple
from jsbgym.assessors import AssessorImpl, ContinuousSequentialAssessor
from jsbgym.tests import stubs as stubs


class TestAssessorImpl(unittest.TestCase):
    def setUp(self):
        pass

    def get_class_under_test(self):
        return AssessorImpl

    def get_assessor(self, *args, **kwargs):
        assessor_class = self.get_class_under_test()
        return assessor_class(*args, **kwargs)

    @staticmethod
    def get_dummy_state_class() -> Type[NamedTuple]:
        return collections.namedtuple("State", ["test_prop1", "test_prop2"])

    def get_dummy_state(self, value1: float = 0.0, value2: float = 1.0):
        State = self.get_dummy_state_class()
        return State(value1, value2)

    def test_init_throws_error_on_empty_base_components(self):
        base_components = ()

        with self.assertRaises(ValueError):
            _ = self.get_assessor(base_components)

    def test_init_throws_error_on_empty_base_components_non_empty_shaping_components(
        self,
    ):
        base_components = ()
        shaping_components = (stubs.ConstantRewardComponentStub(),)

        with self.assertRaises(ValueError):
            _ = self.get_assessor(base_components, shaping_components)

    def test_calculate_single_base_component(self):
        for positive_rewards in (True, False):
            component = stubs.ConstantRewardComponentStub()
            assessor = self.get_assessor(
                base_components=(component,), positive_rewards=positive_rewards
            )
            state = self.get_dummy_state()

            reward = assessor.assess(state, state, True)

            if positive_rewards:
                expected_shaping_reward = component.get_return_value()
            else:
                expected_shaping_reward = 1 - component.get_return_value()

            # should be same because not shaping
            expected_non_shaping_reward = expected_shaping_reward
            self.assertAlmostEqual(expected_shaping_reward, reward.agent_reward())
            self.assertAlmostEqual(
                expected_non_shaping_reward, reward.assessment_reward()
            )

    def test_calculate_multiple_base_components(self):
        for positive_rewards in (True, False):
            reward_values = [0.1, 0.2, 0.4]
            components = tuple(
                stubs.ConstantRewardComponentStub(val) for val in reward_values
            )
            assessor = self.get_assessor(components, positive_rewards=positive_rewards)
            state = self.get_dummy_state_class()

            reward = assessor.assess(state, state, True)

            if positive_rewards:
                comp_values = [cmp.get_return_value() for cmp in components]
            else:
                comp_values = [cmp.get_return_value() - 1 for cmp in components]

            expected_shaping_reward = sum(comp_values) / len(comp_values)
            # should be same because not shaping
            expected_non_shaping_reward = expected_shaping_reward
            self.assertAlmostEqual(
                expected_shaping_reward,
                reward.agent_reward(),
                msg=f"positive reward {positive_rewards}",
            )
            self.assertAlmostEqual(
                expected_non_shaping_reward,
                reward.assessment_reward(),
                msg=f"positive reward {positive_rewards}",
            )

    def test_calculate_with_shaping_components(self):
        for positive_rewards in (True, False):
            base_reward_vals = [0.0, 0.1]
            shaping_reward_vals = [0.2, 0.3]
            base_components = tuple(
                stubs.ConstantRewardComponentStub(val) for val in base_reward_vals
            )
            shape_components = tuple(
                stubs.ConstantRewardComponentStub(val) for val in shaping_reward_vals
            )
            assessor = self.get_assessor(
                base_components, shape_components, positive_rewards=positive_rewards
            )
            state = self.get_dummy_state()

            reward = assessor.assess(state, state, True)

            if positive_rewards:
                expected_shaping_reward = sum(
                    base_reward_vals + shaping_reward_vals
                ) / len(base_reward_vals + shaping_reward_vals)
                expected_non_shaping_reward = sum(base_reward_vals) / len(
                    base_reward_vals
                )
            else:
                base_negative_vals = list(base_val - 1 for base_val in base_reward_vals)
                expected_shaping_reward = sum(
                    base_negative_vals + shaping_reward_vals
                ) / len(base_negative_vals + shaping_reward_vals)
                expected_non_shaping_reward = sum(base_negative_vals) / len(
                    base_negative_vals
                )

            self.assertAlmostEqual(
                expected_shaping_reward,
                reward.agent_reward(),
                msg=f"positive reward {positive_rewards}",
            )
            self.assertAlmostEqual(
                expected_non_shaping_reward,
                reward.assessment_reward(),
                msg=f"positive reward {positive_rewards}",
            )


class TestContinuousSequentialAssessor(TestAssessorImpl):
    def get_class_under_test(self):
        return ContinuousSequentialAssessor

    def test_calculate_with_shaping_components(self):
        for positive_rewards in (True, False):
            num_state_vars = 3
            (
                DummyState,
                props,
            ) = stubs.FlightTaskStub.get_dummy_state_class_and_properties(
                num_state_vars
            )
            base_reward = 0
            base_component = stubs.ConstantRewardComponentStub(0)

            # create two states with a component that will recognise as low and high potential resp.
            state_low_potential = DummyState(*(1.0 for _ in range(num_state_vars)))
            state_high_potential = DummyState(*(2.0 for _ in range(num_state_vars)))
            low_potential, high_potential = 0.5, 1.0
            potential_map = {
                state_low_potential: low_potential,
                state_high_potential: high_potential,
            }
            shape_component = stubs.RewardComponentStub(potential_map)

            assessor = self.get_assessor(
                (base_component,), (shape_component,), positive_rewards=positive_rewards
            )

            # if non-terminal, expect to see reward equal to potential increase
            terminal = False
            reward = assessor.assess(
                state_high_potential, state_low_potential, terminal
            )

            base_reward_as_configured = (
                base_reward if positive_rewards else base_reward - 1
            )
            expected_shaping_reward = (
                base_reward_as_configured + (high_potential - low_potential)
            ) / 2
            expected_non_shaping_reward = base_reward_as_configured

            msg = f"positive reward {positive_rewards}"
            self.assertAlmostEqual(
                expected_shaping_reward, reward.agent_reward(), msg=msg
            )
            self.assertAlmostEqual(
                expected_non_shaping_reward, reward.assessment_reward(), msg=msg
            )

            # if terminal, expect to see reward as if terminal step potential was zero
            terminal = True
            terminal_potential = 0.0
            reward = assessor.assess(
                state_high_potential, state_low_potential, terminal
            )

            expected_shaping_reward = (
                base_reward_as_configured + (terminal_potential - low_potential)
            ) / 2
            expected_non_shaping_reward = base_reward_as_configured

            self.assertAlmostEqual(
                expected_shaping_reward, reward.agent_reward(), msg=msg
            )
            self.assertAlmostEqual(
                expected_non_shaping_reward, reward.assessment_reward(), msg=msg
            )

    def assess_reward_for_potential_change_with_dependency(
        self,
        state_potential: float,
        prev_state_potential: float,
        dependency_potential: float,
        positive_rewards: bool,
    ):
        """
        Calculates the reward given we transition from prev_state_potential to
        state_potential, and a dependant component unchanged at dependency_potential.

        Step is non-terminal and base reward is zero.
        """
        num_state_vars = 3
        terminal = False

        DummyState, props = stubs.FlightTaskStub.get_dummy_state_class_and_properties(
            num_state_vars
        )

        # want base reward to be zero so we can focus on shaped reward
        if positive_rewards:
            base_component = stubs.ConstantRewardComponentStub(0)
        else:
            base_component = stubs.ConstantRewardComponentStub(1)
        # create two states with a component that will recognise as low and high potential resp.
        state = DummyState(*(1.0 for _ in range(num_state_vars)))
        prev_state = DummyState(*(2.0 for _ in range(num_state_vars)))
        potential_map = {state: state_potential, prev_state: prev_state_potential}

        # make components
        shape_component = stubs.RewardComponentStub(potential_map)
        dependency_potential_map = {key: dependency_potential for key in potential_map}
        dependant_shape_component = stubs.RewardComponentStub(dependency_potential_map)

        dependency_map = {shape_component: (dependant_shape_component,)}

        assessor = self.get_assessor(
            (base_component,),
            (shape_component, dependant_shape_component),
            potential_dependency_map=dependency_map,
            positive_rewards=positive_rewards,
        )

        return assessor.assess(state, prev_state, terminal)

    def test_calculate_with_shaping_components_and_dependency_at_zero(self):
        for positive_rewards in (True, False):
            low_potential = 0.5
            high_potential = 1.0
            dependent_potential = 0.0
            reward = self.assess_reward_for_potential_change_with_dependency(
                low_potential, high_potential, dependent_potential, positive_rewards
            )

            # we expect to have had a potential improvement, but because dependent component is at
            #   zero potential, no reward is given
            expected_reward = 0.0

            self.assertAlmostEqual(expected_reward, reward.agent_reward())

    def test_calculate_with_shaping_components_and_dependency_at_fraction(self):
        for positive_rewards in (True, False):
            low_potential = 0.5
            high_potential = 1.0
            dependent_potential = 0.75
            reward = self.assess_reward_for_potential_change_with_dependency(
                high_potential, low_potential, dependent_potential, positive_rewards
            )

            total_reward_values = (high_potential - low_potential) * dependent_potential
            # there are 3 components (incl. the dependant component and base
            #   component with value 0) so divide
            averaged_reward_values = total_reward_values / (
                len(reward.shaping_reward_elements) + len(reward.base_reward_elements)
            )

            self.assertAlmostEqual(averaged_reward_values, reward.agent_reward())

    def test_calculate_with_shaping_components_and_dependency_at_one(self):
        for positive_rewards in (True, False):
            low_potential = 0.5
            high_potential = 1.0
            dependent_potential = 1.0
            reward = self.assess_reward_for_potential_change_with_dependency(
                high_potential, low_potential, dependent_potential, positive_rewards
            )

            total_reward_values = (high_potential - low_potential) * dependent_potential
            # there are 3 components (incl. the dependant component and base
            #   component with value 0) so divide
            averaged_reward_values = total_reward_values / (
                len(reward.shaping_reward_elements) + len(reward.base_reward_elements)
            )

            self.assertAlmostEqual(averaged_reward_values, reward.agent_reward())



import unittest
import numpy as np
from jsbgym.agents import RandomAgent, ConstantAgent
from jsbgym.tests.stubs import FlightTaskStub


class TestRandomAgent(unittest.TestCase):
    def setUp(self):
        self.action_space = FlightTaskStub().get_action_space()
        self.agent = RandomAgent(action_space=self.action_space)

    def test_act_generates_valid_actions(self):
        num_test_actions = 5
        for _ in range(num_test_actions):
            action = self.agent.act(None)
            self.assertTrue(self.action_space.contains(action))


class TestConstantAgent(unittest.TestCase):
    def setUp(self):
        self.task = FlightTaskStub()
        self.action_space = self.task.get_action_space()
        self.agent = ConstantAgent(action_space=self.action_space)

    def test_act_generates_valid_actions(self):
        num_test_actions = 3
        for _ in range(num_test_actions):
            action = self.agent.act(None)
            self.assertTrue(self.action_space.contains(action))

    def test_act_returns_same_action(self):
        num_test_actions = 5
        old_action = self.agent.act(None)
        for _ in range(num_test_actions):
            action = self.agent.act(None)
            np.testing.assert_array_almost_equal(old_action, action)
            old_action = action



import collections


class Aircraft(
    collections.namedtuple(
        "Aircraft", ["jsbsim_id", "flightgear_id", "name", "cruise_speed_kts"]
    )
):
    KTS_TO_M_PER_S = 0.51444
    KTS_TO_FT_PER_S = 1.6878

    def get_max_distance_m(self, episode_time_s: float) -> float:
        """Estimates the maximum distance this aircraft can travel in an episode"""
        margin = 0.1
        return (
            self.cruise_speed_kts * self.KTS_TO_M_PER_S * episode_time_s * (1 + margin)
        )

    def get_cruise_speed_fps(self) -> float:
        return self.cruise_speed_kts * self.KTS_TO_FT_PER_S


cessna172P = Aircraft("c172p", "c172p", "Cessna172P", 120)
pa28 = Aircraft("pa28", "PA28-161-180", "PA28", 130)
j3 = Aircraft("J3Cub", "J3Cub", "J3", 70)
f15 = Aircraft("f15", "f15c", "F15", 500)
f16 = Aircraft("f16", "f16-block-52", "F16", 550)
ov10 = Aircraft("OV10", "OV10_USAFE", "OV10", 200)
a320 = Aircraft("A320", "A320-200-CFM", "A320", 250)
b747 = Aircraft("B747", "747-400", "B747", 250)
md11 = Aircraft("MD11", "MD-11", "MD11", 250)



import functools
import operator
from typing import Tuple
from jsbgym.aircraft import cessna172P, a320, f15, pa28, b747, f16, j3, md11, ov10
from typing import Dict, Iterable


class AttributeFormatter(object):
    """
    Replaces characters that would be illegal in an attribute name

    Used through its static method, translate()
    """

    ILLEGAL_CHARS = "\-/."
    TRANSLATE_TO = "_" * len(ILLEGAL_CHARS)
    TRANSLATION_TABLE = str.maketrans(ILLEGAL_CHARS, TRANSLATE_TO)

    @staticmethod
    def translate(string: str):
        return string.translate(AttributeFormatter.TRANSLATION_TABLE)


def get_env_id(aircraft, task_type, shaping, enable_flightgear) -> str:
    """
    Creates an env ID from the environment's components
    :param task_type: Task class, the environment's task
    :param aircraft: Aircraft namedtuple, the aircraft to be flown
    :param shaping: HeadingControlTask.Shaping enum, the reward shaping setting
    :param enable_flightgear: True if FlightGear simulator is enabled for visualisation else False
    """
    if enable_flightgear:
        fg_setting = "FG"
    else:
        fg_setting = "NoFG"
    return f"{aircraft.name}-{task_type.__name__}-{shaping}-{fg_setting}-v0"


def get_env_id_kwargs_map() -> Dict[str, Tuple]:
    """Returns all environment IDs mapped to tuple of (task, aircraft, shaping, flightgear)"""
    # lazy import to avoid circular dependencies
    from jsbgym.tasks import Shaping, HeadingControlTask, TurnHeadingControlTask

    map = {}
    for task_type in (HeadingControlTask, TurnHeadingControlTask):
        for plane in (cessna172P, a320, f15, pa28, b747, f16, j3, md11, ov10):
            for shaping in (Shaping.STANDARD, Shaping.EXTRA, Shaping.EXTRA_SEQUENTIAL):
                for enable_flightgear in (True, False):
                    id = get_env_id(plane, task_type, shaping, enable_flightgear)
                    assert id not in map
                    map[id] = (plane, task_type, shaping, enable_flightgear)
    return map


def product(iterable: Iterable):
    """
    Multiplies all elements of iterable and returns result

    ATTRIBUTION: code provided by Raymond Hettinger on SO
    https://stackoverflow.com/questions/595374/whats-the-function-like-sum-but-for-multiplication-product
    """
    return functools.reduce(operator.mul, iterable, 1)


def reduce_reflex_angle_deg(angle: float) -> float:
    """Given an angle in degrees, normalises in [-179, 180]"""
    # ATTRIBUTION: solution from James Polk on SO,
    # https://stackoverflow.com/questions/2320986/easy-way-to-keeping-angles-between-179-and-180-degrees#
    new_angle = angle % 360
    if new_angle > 180:
        new_angle -= 360
    return new_angle



import jsbsim
import os
import time
from mpl_toolkits.mplot3d import Axes3D  # required for 3d plotting
from typing import Dict, Union
import jsbgym.properties as prp
from jsbgym.aircraft import Aircraft, cessna172P


class Simulation(object):
    """
    A class which wraps an instance of JSBSim and manages communication with it.
    """

    encoding = "utf-8"  # encoding of bytes returned by JSBSim Cython funcs
    ROOT_DIR = os.path.abspath(jsbsim.__path__[0])
    OUTPUT_FILE = "flightgear.xml"
    LONGITUDINAL = "longitudinal"
    FULL = "full"

    def __init__(
        self,
        sim_frequency_hz: float = 60.0,
        aircraft: Aircraft = cessna172P,
        init_conditions: Dict[prp.Property, float] = None,
        allow_flightgear_output: bool = True,
    ):
        """
        Constructor. Creates an instance of JSBSim and sets initial conditions.

        :param sim_frequency_hz: the JSBSim integration frequency in Hz.
        :param aircraft_model_name: name of aircraft to be loaded.
            JSBSim looks for file \model_name\model_name.xml from root dir.
        :param init_conditions: dict mapping properties to their initial values.
            Defaults to None, causing a default set of initial props to be used.
        :param allow_flightgear_output: bool, loads a config file instructing
            JSBSim to connect to an output socket if True.
        """
        self.jsbsim = jsbsim.FGFDMExec(root_dir=self.ROOT_DIR)
        self.jsbsim.set_debug_level(0)
        if allow_flightgear_output:
            flightgear_output_config = os.path.join(
                os.path.dirname(os.path.abspath(__file__)), self.OUTPUT_FILE
            )
            self.jsbsim.set_output_directive(flightgear_output_config)
        self.sim_dt = 1.0 / sim_frequency_hz
        self.aircraft = aircraft
        self.initialise(self.sim_dt, self.aircraft.jsbsim_id, init_conditions)
        self.jsbsim.disable_output()
        self.wall_clock_dt = None

    def __getitem__(self, prop: Union[prp.BoundedProperty, prp.Property]) -> float:
        """
        Retrieves specified simulation property.

        Properties are identified by strings. A list can be found in the JSBSim
        reference manual, launching JSBSim with '--catalog' command line arg or
        calling FGFDMExec.get_property_catalog().

        :param prop: BoundedProperty, the property to be retrieved
        :return: float
        """
        return self.jsbsim[prop.name]

    def __setitem__(
        self, prop: Union[prp.BoundedProperty, prp.Property], value
    ) -> None:
        """
        Sets simulation property to specified value.

        Properties are identified by strings. A list can be found in the JSBSim
        reference manual, launching JSBSim with '--catalog' command line arg or
        calling FGFDMExec.get_property_catalog().

        Warning: JSBSim will create new properties if the specified one exists.
        If the property you are setting is read-only in JSBSim the operation
        will silently fail.

        :param prop: BoundedProperty, the property to be retrieved
        :param value: object?, the value to be set
        """
        self.jsbsim[prop.name] = value

    def load_model(self, model_name: str) -> None:
        """
        Loads the specified aircraft config into the simulation.

        The root JSBSim directory aircraft folder is searched for the aircraft
        XML config file.

        :param model_name: string, the aircraft name
        """
        load_success = self.jsbsim.load_model(model_name)

        if not load_success:
            raise RuntimeError(
                "JSBSim could not find specified model_name: " + model_name
            )

    def get_aircraft(self) -> Aircraft:
        """
        Gets the Aircraft this sim was initialised with.
        """
        return self.aircraft

    def get_loaded_model_name(self) -> str:
        """
        Gets the name of the aircraft model currently loaded in JSBSim.

        :return: string, the name of the aircraft model if one is loaded, or
            None if no model is loaded.
        """
        name: str = self.jsbsim.get_model_name().decode(self.encoding)
        if name:
            return name
        else:
            # name is empty string, no model is loaded
            return None

    def get_sim_time(self) -> float:
        """Gets the simulation time from JSBSim, a float."""
        return self.jsbsim["simulation/sim-time-sec"]

    def initialise(
        self,
        dt: float,
        model_name: str,
        init_conditions: Dict["prp.Property", float] = None,
    ) -> None:
        """
        Loads an aircraft and initialises simulation conditions.

        JSBSim creates an InitialConditions object internally when given an
        XML config file. This method either loads a basic set of ICs, or
        can be passed a dictionary with ICs. In the latter case a minimal IC
        XML file is loaded, and then the dictionary values are fed in.

        :param dt: float, the JSBSim integration timestep in seconds
        :param model_name: string, name of aircraft to be loaded
        :param init_conditions: dict mapping properties to their initial values
        """
        if init_conditions is not None:
            # if we are specifying conditions, load a minimal file
            ic_file = "minimal_ic.xml"
        else:
            ic_file = "basic_ic.xml"

        ic_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), ic_file)
        self.jsbsim.load_ic(ic_path, useStoredPath=False)
        self.load_model(model_name)
        self.jsbsim.set_dt(dt)
        # extract set of legal property names for this aircraft
        # TODO: can translate the .split(" ")[0] once JSBSim bug has been fixed (in progress)

        # now that IC object is created in JSBSim, specify own conditions
        self.set_custom_initial_conditions(init_conditions)

        success = self.jsbsim.run_ic()
        if not success:
            raise RuntimeError("JSBSim failed to init simulation conditions.")

    def set_custom_initial_conditions(
        self, init_conditions: Dict["prp.Property", float] = None
    ) -> None:
        if init_conditions is not None:
            for prop, value in init_conditions.items():
                self[prop] = value

    def reinitialise(self, init_conditions: Dict["prp.Property", float] = None) -> None:
        """
        Resets JSBSim to initial conditions.

        The same aircraft and other settings are kept loaded in JSBSim. If a
        dict of ICs is provided, JSBSim is initialised using these, else the
        last specified ICs are used.

        :param init_conditions: dict mapping properties to their initial values
        """
        self.set_custom_initial_conditions(init_conditions=init_conditions)
        no_output_reset_mode = 0
        self.jsbsim.reset_to_initial_conditions(no_output_reset_mode)

    def run(self) -> bool:
        """
        Runs a single timestep in the JSBSim simulation.

        JSBSim monitors the simulation and detects whether it thinks it should
        end, e.g. because a simulation time was specified. False is returned
        if JSBSim termination criteria are met.

        :return: bool, False if sim has met JSBSim termination criteria else True.
        """
        result = self.jsbsim.run()
        if self.wall_clock_dt is not None:
            time.sleep(self.wall_clock_dt)
        return result

    def enable_flightgear_output(self):
        self.jsbsim.enable_output()

    def disable_flightgear_output(self):
        self.jsbsim.disable_output()

    def close(self):
        """Closes the simulation and any plots."""
        if self.jsbsim:
            self.jsbsim = None

    def set_simulation_time_factor(self, time_factor):
        """
        Specifies a factor, relative to realtime, for simulation to run at.

        The simulation runs at realtime for time_factor = 1. It runs at double
        speed for time_factor=2, and half speed for 0.5.

        :param time_factor: int or float, nonzero, sim speed relative to realtime
            if None, the simulation is run at maximum computational speed
        """
        if time_factor is None:
            self.wall_clock_dt = None
        elif time_factor <= 0:
            raise ValueError("time factor must be positive and non-zero")
        else:
            self.wall_clock_dt = self.sim_dt / time_factor

    def start_engines(self):
        """Sets all engines running."""
        self[prp.all_engine_running] = -1

    def set_throttle_mixture_controls(self, throttle_cmd: float, mixture_cmd: float):
        """
        Sets throttle and mixture settings

        If an aircraft is multi-engine and has multiple throttle_cmd and mixture_cmd
        controls, sets all of them. Currently only supports up to two throttles/mixtures.
        """
        self[prp.throttle_cmd] = throttle_cmd
        self[prp.mixture_cmd] = mixture_cmd

        try:
            self[prp.throttle_1_cmd] = throttle_cmd
            self[prp.mixture_1_cmd] = mixture_cmd
        except KeyError:
            pass  # must be single-control aircraft

    def raise_landing_gear(self):
        """Raises all aircraft landing gear."""
        self[prp.gear] = 0.0
        self[prp.gear_all_cmd] = 0.0



import gymnasium as gym
import numpy as np
import random
import types
import math
import enum
import warnings
from collections import namedtuple
import jsbgym.properties as prp
from jsbgym import assessors, rewards, utils
from jsbgym.simulation import Simulation
from jsbgym.properties import BoundedProperty, Property
from jsbgym.aircraft import Aircraft
from jsbgym.rewards import RewardStub
from abc import ABC, abstractmethod
from typing import Optional, Sequence, Dict, Tuple, NamedTuple, Type


class Task(ABC):
    """
    Interface for Tasks, modules implementing specific environments in JSBSim.

    A task defines its own state space, action space, termination conditions and agent_reward function.
    """

    @abstractmethod
    def task_step(
        self, sim: Simulation, action: Sequence[float], sim_steps: int
    ) -> Tuple[np.ndarray, float, bool, Dict]:
        """
        Calculates new state, reward and termination.

        :param sim: a Simulation, the simulation from which to extract state
        :param action: sequence of floats, the agent's last action
        :param sim_steps: number of JSBSim integration steps to perform following action
            prior to making observation
        :return: tuple of (observation, reward, terminated, truncated, info) where,
            observation: array, agent's observation of the environment state
            reward: float, the reward for that step
            terminated: bool, True if the episode is over else False
            info: dict, optional, containing diagnostic info for debugging etc.
        """

    ...

    @abstractmethod
    def observe_first_state(self, sim: Simulation) -> np.ndarray:
        """
        Initialise any state/controls and get first state observation from reset sim.

        :param sim: Simulation, the environment simulation
        :return: np array, the first state observation of the episode
        """
        ...

    @abstractmethod
    def get_initial_conditions(self) -> Optional[Dict[Property, float]]:
        """
        Returns dictionary mapping initial episode conditions to values.

        Episode initial conditions (ICs) are defined by specifying values for
        JSBSim properties, represented by their name (string) in JSBSim.

        JSBSim uses a distinct set of properties for ICs, beginning with 'ic/'
        which differ from property names during the simulation, e.g. "ic/u-fps"
        instead of "velocities/u-fps". See https://jsbsim-team.github.io/jsbsim/

        :return: dict mapping string for each initial condition property to
            initial value, a float, or None to use Env defaults
        """
        ...

    @abstractmethod
    def get_state_space(self) -> gym.Space:
        """Get the task's state Space object"""
        ...

    @abstractmethod
    def get_action_space(self) -> gym.Space:
        """Get the task's action Space object"""
        ...


class FlightTask(Task, ABC):
    """
    Abstract superclass for flight tasks.

    Concrete subclasses should implement the following:
        state_variables attribute: tuple of Propertys, the task's state representation
        action_variables attribute: tuple of Propertys, the task's actions
        get_initial_conditions(): returns dict mapping InitialPropertys to initial values
        _is_terminal(): determines episode termination
        (optional) _new_episode_init(): performs any control input/initialisation on episode reset
        (optional) _update_custom_properties: updates any custom properties in the sim
    """

    INITIAL_ALTITUDE_FT = 5000
    base_state_variables = (
        prp.altitude_sl_ft,
        prp.pitch_rad,
        prp.roll_rad,
        prp.u_fps,
        prp.v_fps,
        prp.w_fps,
        prp.p_radps,
        prp.q_radps,
        prp.r_radps,
        prp.aileron_left,
        prp.aileron_right,
        prp.elevator,
        prp.rudder,
    )
    base_initial_conditions = (
        types.MappingProxyType(  # MappingProxyType makes dict immutable
            {
                prp.initial_altitude_ft: INITIAL_ALTITUDE_FT,
                prp.initial_terrain_altitude_ft: 0.00000001,
                prp.initial_longitude_geoc_deg: -2.3273,
                prp.initial_latitude_geod_deg: 51.3781,  # corresponds to UoBath
            }
        )
    )
    last_agent_reward = Property(
        "reward/last_agent_reward",
        "agent reward from step; includes potential-based shaping reward",
    )
    last_assessment_reward = Property(
        "reward/last_assess_reward", "assessment reward from step; excludes shaping"
    )
    state_variables: Tuple[BoundedProperty, ...]
    action_variables: Tuple[BoundedProperty, ...]
    assessor: assessors.Assessor
    State: Type[NamedTuple]

    def __init__(self, assessor: assessors.Assessor, debug: bool = False) -> None:
        self.last_state = None
        self.assessor = assessor
        self._make_state_class()
        self.debug = debug

    def _make_state_class(self) -> None:
        """Creates a namedtuple for readable State data"""
        # get list of state property names, containing legal chars only
        legal_attribute_names = [prop.get_legal_name() for prop in self.state_variables]
        self.State = namedtuple("State", legal_attribute_names)

    def task_step(
        self, sim: Simulation, action: Sequence[float], sim_steps: int
    ) -> Tuple[NamedTuple, float, bool, Dict]:
        # input actions
        for prop, command in zip(self.action_variables, action):
            sim[prop] = command

        # run simulation
        for _ in range(sim_steps):
            sim.run()

        self._update_custom_properties(sim)
        state = self.State(*(sim[prop] for prop in self.state_variables))
        terminated = self._is_terminal(sim)
        truncated = False
        reward = self.assessor.assess(state, self.last_state, terminated)
        if terminated:
            reward = self._reward_terminal_override(reward, sim)
        if self.debug:
            self._validate_state(state, terminated, truncated, action, reward)
        self._store_reward(reward, sim)
        self.last_state = state
        info = {"reward": reward}

        return state, reward.agent_reward(), terminated, False, info

    def _validate_state(self, state, terminated, truncated, action, reward):
        if any(math.isnan(el) for el in state):  # float('nan') in state doesn't work!
            msg = (
                f"Invalid state encountered!\n"
                f"State: {state}\n"
                f"Prev. State: {self.last_state}\n"
                f"Action: {action}\n"
                f"Terminal: {terminated}\n"
                f"Trucantion: {truncated}\n"
                f"Reward: {reward}"
            )
            warnings.warn(msg, RuntimeWarning)

    def _store_reward(self, reward: rewards.Reward, sim: Simulation):
        sim[self.last_agent_reward] = reward.agent_reward()
        sim[self.last_assessment_reward] = reward.assessment_reward()

    def _update_custom_properties(self, sim: Simulation) -> None:
        """Calculates any custom properties which change every timestep."""
        pass

    @abstractmethod
    def _is_terminal(self, sim: Simulation) -> bool:
        """Determines whether the current episode should terminate.

        :param sim: the current simulation
        :return: True if the episode should terminate else False
        """
        ...

    @abstractmethod
    def _reward_terminal_override(
        self, reward: rewards.Reward, sim: Simulation
    ) -> bool:
        """
        Determines whether a custom reward is needed, e.g. because
        a terminal condition is met.
        """
        ...

    def observe_first_state(self, sim: Simulation) -> np.ndarray:
        self._new_episode_init(sim)
        self._update_custom_properties(sim)
        state = self.State(*(sim[prop] for prop in self.state_variables))
        self.last_state = state
        return state

    def _new_episode_init(self, sim: Simulation) -> None:
        """
        This method is called at the start of every episode. It is used to set
        the value of any controls or environment properties not already defined
        in the task's initial conditions.

        By default it simply starts the aircraft engines.
        """
        sim.start_engines()
        sim.raise_landing_gear()
        self._store_reward(RewardStub(1.0, 1.0), sim)

    @abstractmethod
    def get_initial_conditions(self) -> Dict[Property, float]:
        ...

    def get_state_space(self) -> gym.Space:
        state_lows = np.array([state_var.min for state_var in self.state_variables])
        state_highs = np.array([state_var.max for state_var in self.state_variables])
        return gym.spaces.Box(low=state_lows, high=state_highs, dtype=np.float64)

    def get_action_space(self) -> gym.Space:
        action_lows = np.array([act_var.min for act_var in self.action_variables])
        action_highs = np.array([act_var.max for act_var in self.action_variables])
        return gym.spaces.Box(low=action_lows, high=action_highs, dtype=np.float64)


class Shaping(enum.Enum):
    STANDARD = "STANDARD"
    EXTRA = "EXTRA"
    EXTRA_SEQUENTIAL = "EXTRA_SEQUENTIAL"


class HeadingControlTask(FlightTask):
    """
    A task in which the agent must perform steady, level flight maintaining its
    initial heading.
    """

    THROTTLE_CMD = 0.8
    MIXTURE_CMD = 0.8
    INITIAL_HEADING_DEG = 270
    DEFAULT_EPISODE_TIME_S = 60.0
    ALTITUDE_SCALING_FT = 150
    TRACK_ERROR_SCALING_DEG = 8
    ROLL_ERROR_SCALING_RAD = 0.15  # approx. 8 deg
    SIDESLIP_ERROR_SCALING_DEG = 3.0
    MIN_STATE_QUALITY = 0.0  # terminate if state 'quality' is less than this
    MAX_ALTITUDE_DEVIATION_FT = 1000  # terminate if altitude error exceeds this
    target_track_deg = BoundedProperty(
        "target/track-deg",
        "desired heading [deg]",
        prp.heading_deg.min,
        prp.heading_deg.max,
    )
    track_error_deg = BoundedProperty(
        "error/track-error-deg", "error to desired track [deg]", -180, 180
    )
    altitude_error_ft = BoundedProperty(
        "error/altitude-error-ft",
        "error to desired altitude [ft]",
        prp.altitude_sl_ft.min,
        prp.altitude_sl_ft.max,
    )
    action_variables = (prp.aileron_cmd, prp.elevator_cmd, prp.rudder_cmd)

    def __init__(
        self,
        shaping_type: Shaping,
        step_frequency_hz: float,
        aircraft: Aircraft,
        episode_time_s: float = DEFAULT_EPISODE_TIME_S,
        positive_rewards: bool = True,
    ):
        """
        Constructor.

        :param step_frequency_hz: the number of agent interaction steps per second
        :param aircraft: the aircraft used in the simulation
        """
        self.max_time_s = episode_time_s
        episode_steps = math.ceil(self.max_time_s * step_frequency_hz)
        self.steps_left = BoundedProperty(
            "info/steps_left", "steps remaining in episode", 0, episode_steps
        )
        self.aircraft = aircraft
        self.extra_state_variables = (
            self.altitude_error_ft,
            prp.sideslip_deg,
            self.track_error_deg,
            self.steps_left,
        )
        self.state_variables = (
            FlightTask.base_state_variables + self.extra_state_variables
        )
        self.positive_rewards = positive_rewards
        assessor = self.make_assessor(shaping_type)
        super().__init__(assessor)

    def make_assessor(self, shaping: Shaping) -> assessors.AssessorImpl:
        base_components = self._make_base_reward_components()
        shaping_components = ()
        return self._select_assessor(base_components, shaping_components, shaping)

    def _make_base_reward_components(self) -> Tuple[rewards.RewardComponent, ...]:
        base_components = (
            rewards.AsymptoticErrorComponent(
                name="altitude_error",
                prop=self.altitude_error_ft,
                state_variables=self.state_variables,
                target=0.0,
                is_potential_based=False,
                scaling_factor=self.ALTITUDE_SCALING_FT,
            ),
            rewards.AsymptoticErrorComponent(
                name="travel_direction",
                prop=self.track_error_deg,
                state_variables=self.state_variables,
                target=0.0,
                is_potential_based=False,
                scaling_factor=self.TRACK_ERROR_SCALING_DEG,
            ),
            # add an airspeed error relative to cruise speed component?
        )
        return base_components

    def _select_assessor(
        self,
        base_components: Tuple[rewards.RewardComponent, ...],
        shaping_components: Tuple[rewards.RewardComponent, ...],
        shaping: Shaping,
    ) -> assessors.AssessorImpl:
        if shaping is Shaping.STANDARD:
            return assessors.AssessorImpl(
                base_components,
                shaping_components,
                positive_rewards=self.positive_rewards,
            )
        else:
            wings_level = rewards.AsymptoticErrorComponent(
                name="wings_level",
                prop=prp.roll_rad,
                state_variables=self.state_variables,
                target=0.0,
                is_potential_based=True,
                scaling_factor=self.ROLL_ERROR_SCALING_RAD,
            )
            no_sideslip = rewards.AsymptoticErrorComponent(
                name="no_sideslip",
                prop=prp.sideslip_deg,
                state_variables=self.state_variables,
                target=0.0,
                is_potential_based=True,
                scaling_factor=self.SIDESLIP_ERROR_SCALING_DEG,
            )
            potential_based_components = (wings_level, no_sideslip)

        if shaping is Shaping.EXTRA:
            return assessors.AssessorImpl(
                base_components,
                potential_based_components,
                positive_rewards=self.positive_rewards,
            )
        elif shaping is Shaping.EXTRA_SEQUENTIAL:
            altitude_error, travel_direction = base_components
            # make the wings_level shaping reward dependent on facing the correct direction
            dependency_map = {wings_level: (travel_direction,)}
            return assessors.ContinuousSequentialAssessor(
                base_components,
                potential_based_components,
                potential_dependency_map=dependency_map,
                positive_rewards=self.positive_rewards,
            )

    def get_initial_conditions(self) -> Dict[Property, float]:
        extra_conditions = {
            prp.initial_u_fps: self.aircraft.get_cruise_speed_fps(),
            prp.initial_v_fps: 0,
            prp.initial_w_fps: 0,
            prp.initial_p_radps: 0,
            prp.initial_q_radps: 0,
            prp.initial_r_radps: 0,
            prp.initial_roc_fpm: 0,
            prp.initial_heading_deg: self.INITIAL_HEADING_DEG,
        }
        return {**self.base_initial_conditions, **extra_conditions}

    def _update_custom_properties(self, sim: Simulation) -> None:
        self._update_track_error(sim)
        self._update_altitude_error(sim)
        self._decrement_steps_left(sim)

    def _update_track_error(self, sim: Simulation):
        v_north_fps, v_east_fps = sim[prp.v_north_fps], sim[prp.v_east_fps]
        track_deg = prp.Vector2(v_east_fps, v_north_fps).heading_deg()
        target_track_deg = sim[self.target_track_deg]
        error_deg = utils.reduce_reflex_angle_deg(track_deg - target_track_deg)
        sim[self.track_error_deg] = error_deg

    def _update_altitude_error(self, sim: Simulation):
        altitude_ft = sim[prp.altitude_sl_ft]
        target_altitude_ft = self._get_target_altitude()
        error_ft = altitude_ft - target_altitude_ft
        sim[self.altitude_error_ft] = error_ft

    def _decrement_steps_left(self, sim: Simulation):
        sim[self.steps_left] -= 1

    def _is_terminal(self, sim: Simulation) -> bool:
        # terminate when time >= max, but use math.isclose() for float equality test
        terminal_step = sim[self.steps_left] <= 0
        state_quality = sim[self.last_assessment_reward]
        # TODO: issues if sequential?
        state_out_of_bounds = state_quality < self.MIN_STATE_QUALITY
        return terminal_step or state_out_of_bounds or self._altitude_out_of_bounds(sim)

    def _altitude_out_of_bounds(self, sim: Simulation) -> bool:
        altitude_error_ft = sim[self.altitude_error_ft]
        return abs(altitude_error_ft) > self.MAX_ALTITUDE_DEVIATION_FT

    def _get_out_of_bounds_reward(self, sim: Simulation) -> rewards.Reward:
        """
        if aircraft is out of bounds, we give the largest possible negative reward:
        as if this timestep, and every remaining timestep in the episode was -1.
        """
        reward_scalar = (1 + sim[self.steps_left]) * -1.0
        return RewardStub(reward_scalar, reward_scalar)

    def _reward_terminal_override(
        self, reward: rewards.Reward, sim: Simulation
    ) -> rewards.Reward:
        if self._altitude_out_of_bounds(sim) and not self.positive_rewards:
            # if using negative rewards, need to give a big negative reward on terminal
            return self._get_out_of_bounds_reward(sim)
        else:
            return reward

    def _new_episode_init(self, sim: Simulation) -> None:
        super()._new_episode_init(sim)
        sim.set_throttle_mixture_controls(self.THROTTLE_CMD, self.MIXTURE_CMD)
        sim[self.steps_left] = self.steps_left.max
        sim[self.target_track_deg] = self._get_target_track()

    def _get_target_track(self) -> float:
        # use the same, initial heading every episode
        return self.INITIAL_HEADING_DEG

    def _get_target_altitude(self) -> float:
        return self.INITIAL_ALTITUDE_FT

    def get_props_to_output(self) -> Tuple:
        return (
            prp.u_fps,
            prp.altitude_sl_ft,
            self.altitude_error_ft,
            self.target_track_deg,
            self.track_error_deg,
            prp.roll_rad,
            prp.sideslip_deg,
            self.last_agent_reward,
            self.last_assessment_reward,
            self.steps_left,
        )


class TurnHeadingControlTask(HeadingControlTask):
    """
    A task in which the agent must make a turn from a random initial heading,
    and fly level to a random target heading.
    """

    def get_initial_conditions(self) -> [Dict[Property, float]]:
        initial_conditions = super().get_initial_conditions()
        random_heading = random.uniform(prp.heading_deg.min, prp.heading_deg.max)
        initial_conditions[prp.initial_heading_deg] = random_heading
        return initial_conditions

    def _get_target_track(self) -> float:
        # select a random heading each episode
        return random.uniform(self.target_track_deg.min, self.target_track_deg.max)
